Package: akin
Type: Package
Title: Functional Utilities for Data Processing
Version: 0.3.2
Authors@R: c(
    person(
      "Dragos", "Bandur",
      email = "dbandur@sympatico.ca",
      role = c("aut", "cre")
    )
  )
Description: Covers several areas of data processing: batch-splitting, reading and writing of large
    data files, data tiling, one-hot encoding and decoding of data tiles, stratified proportional
    (random or probabilistic) data sampling, data normalization and thresholding, substring location
    and commonalities inside strings, and location and tabulation of amino acids, modifications or 
    associated monoisotopic masses inside modified peptides. The extractor utility implements code from
    'Matrix.utils', Varrichio C (2020), <https://cran.r-project.org/package=Matrix.utils>.
License: GPL (>= 3)
Imports: data.table (>= 1.18.2.0), RVerbalExpressions (>= 0.1.0),
        RcppAlgos(>= 2.9.3), methods
Depends: R (>= 4.1.0), callr, fastmatch, future.apply, listenv, Matrix,
        utils
Encoding: UTF-8
Suggests: future, testthat (>= 3.0.0)
Config/testthat/edition: 3
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-03 18:50:39 UTC; Dragos
Author: Dragos Bandur [aut, cre]
Maintainer: Dragos Bandur <dbandur@sympatico.ca>
Repository: CRAN
Date/Publication: 2026-05-04 04:30:02 UTC
Built: R 4.6.0; ; 2026-05-04 06:44:22 UTC; unix

# Generated by roxygen2: do not edit by hand

export(common)
export(cover)
export(fcommon)
export(findLoc)
export(getEV)
export(locateMod)
export(oneHot)
export(score)
export(splitH)
export(splitV)
export(stratify)
export(tileData)
export(tileHot)
import(RVerbalExpressions)
import(RcppAlgos)
import(data.table)
import(methods)
importFrom(Matrix,Matrix)
importFrom(Matrix,rowSums)
importFrom(Matrix,rsparsematrix)
importFrom(callr,r_bg)
importFrom(fastmatch,fmatch)
importFrom(future.apply,future_lapply)
importFrom(listenv,listenv)
importFrom(stats,median)
importFrom(stats,setNames)
importFrom(utils,getFromNamespace)
importFrom(utils,type.convert)

#' @title One-hot Encoder Of Tiled Data
#'
#' @description One-hot encodes tiled data.
#'
#' @param readpath character, length 1. Path to source data that is readable with [data.table::fread]
#' @param rows integer length 1. Number of rows in each data subset. Internally, it determines the total
#'   number of subsets before the vertical split
#' @param splits integer, length 1. Number of vertical data splits in each subset, see [splitV]. Recommended for
#'   very wide data frames. When \code{splits = 0}, no vertical splitting occurs
#' @param omc character length 1. \strong{O}utput \strong{m}atrix \strong{c}lass. Default, "dgCMatrix". Other option:
#' "matrix"
#' @param ... reserved for [splitH] function arguments, such as \code{dropcols} or \code{orn = TRUE} which is needed for
#' single matrix output
#'
#' @details This utility reads the data in disjoint subsets, tiles them and then one-hot encodes each tile. Encoded
#' tiles are returned as nested list of matrices, as a single matrix, as data frame or as a two-component data frame
#' and sparse matrix list, decided through combinations of \code{dropcols}, \code{omc} and \code{orn} values.
#'
#' NOTE 1: traceability is assured by assembling the data as character names and values from columns marked for encoding.
#' As side effect, at run time the encoding is reported as being applied to "integer(ish)" values only with no loss in
#' accuracy. Empty source data columns gain the "NA" suffix and become single-column, single-valued matrices.
#'
#' NOTE 2: this utility implements background, processing. Check "Security Considerations" in \strong{callr}
#' package documentation.
#'
#' @returns
#'   * When \code{orn = FALSE}, an unnamed \link[listenv]{listenv} of sparse matrices. Recommended for very large
#' source data files. Before proceeding with list output, read NOTE 2 in [getEV] documentation. See Examples 1 and 2.
#'
#'   * When \code{orn = TRUE}, a matrix.
#'
#' NOTE 3: In this case, row and column binding operations were avoided to prevent situations described in NOTE 2, [getEV]
#' documentation. As result, the output matrix is gradually populated instead of being gradually expanded.
#'
#' While \code{orn = TRUE} and \code{dropcols != NULL}:
#'
#'   * When \code{omc = 'matrix'}, a \link[data.table]{data.table} containing encoded, as well as unencoded, dropped
#' columns placed in the leftmost positions.
#'
#'   * When \code{omc = 'dgCMatrix'}, a two-component \link[listenv]{listenv}: a data table containing dropped,
#' unencoded columns and a sparse matrix containing the encoded columns. The row order in both components is identical.
#' See Examples below, and Example 2 in [getEV] documentation.
#'
#' NOTE 4: In all above cases, specific encoded variables can be obtained with [getEV] extractor. When \code{orn = TRUE},
#' [oneHot] \emph{decoded} variables extracted from matrix outputs return named vectors having row numbers as names.
#'
#' @seealso [splitH], [splitV], [oneHot], \link[listenv]{listenv}, \link[Matrix]{Matrix}
#'
#' @keywords conversion
#'
#' @export
#'
#' @examples
#'
#' if (interactive()) {
#'
#' # 1. Shuffled data
#'
#' tempf = tempfile(fileext = '.csv')
#' data(iris)
#' iris22 = iris[{ set.seed(327); sample.int(150) },]          # shuffled iris data
#' rownames(iris22) <- NULL                                    # remove shuffled row names
#' write.table(iris22, tempf, sep = ',', row.names = FALSE, quote = FALSE)
#'
#' # 1.1 Output as List
#' # In most cases, list output requires shuffled data!
#'
#' A = tileHot(readpath = tempf
#'           , rows = 14, splits = 3, print = FALSE)           # encoded data tiles
#' print(A)                                                    # a listenv
#' print(A[[1]])                                               # a snapshot
#'
#' # 1.2 Retrieve iris22 data from encoded list output
#' X = sapply(names(iris22), \(n) getEV(A, n))                 # extract all encoded columns
#' Y = lapply(
#'          lapply(X, oneHot, decode)
#'                                , unname)                    # decoded columns are named vectors!
#' d = as.data.frame(Y)
#' identical(iris22, d)                                        # TRUE
#'
#' unlink(tempf)
#'
#' # 2. Unshuffled data
#'
#' # Make unshuffled data 'csv' file
#' tempf = tempfile(fileext = '.csv')
#' write.table(iris, tempf, sep = ',', row.names = FALSE, quote = FALSE)
#'
#' # 2.1 Output as list
#' # List output fails low cardinality variables on unshuffled data.
#'
#' E = tileHot(readpath = tempf
#'               , rows = 14, splits = 3, print = FALSE)      # same as above
#'
#' # 2.2 Retrieve iris data from encoded list output
#' V = sapply(names(iris), \(n) getEV(E, n))                  # warning
#' W = lapply(
#'          lapply(V, oneHot, decode)
#'                                , unname)                   # decoded columns are named vectors!
#' dd = as.data.frame(W)
#' identical(iris, dd)                                        # FALSE
#' all.equal(iris, dd)                                        # low cardinality "Species"
#'
#' # 2.3 Output as matrix
#' # Matrix output handles low cardinality variables. No data shuffling required.
#'
#' m = tileHot(readpath = tempf                               # low cardinality "Species"
#'           , rows = 14
#'           , splits = 3
#'           , orn = TRUE,                                    # needed for matrix output
#'           , print = FALSE)
#' print(m)                                                   # 150x126 sparse matrix
#'
#' # 2.4 Retrieve iris data from encoded matrix output
#' P = sapply(names(iris), \(n) getEV(m, n))                  # extract encoded columns
#' Q = lapply(
#'            lapply(P, oneHot, decode)
#'                               , unname)                    # decoded columns are named vectors!
#' R = as.data.frame(Q)
#' identical(iris, R)                                         # TRUE
#'
#' # 2.5 Output as "data.table" class
#' D = tileHot(readpath = tempf
#'           , rows = 14
#'           , splits = 3
#'           , omc = 'matrix'                                 # encoded dense matrix
#'           , dropcols = c('Petal.Width', 'Petal.Length')    # unencoded columns
#'           , orn = TRUE                                     # needed for matrix output
#'           , print = FALSE)
#' print(head(D, 10))                                         # a "data.table" class
#' dim(D)                                                     # 150x63
#'
#' # 2.6 Output as a 2-component list
#' Dl = tileHot(readpath = tempf
#'           , rows = 14
#'           , splits = 3
#'           , omc = 'dgCMatrix'                              # the default class
#'           , dropcols = c('Petal.Width', 'Petal.Length')    # unencoded columns
#'           , orn = TRUE                                     # needed for matrix output
#'           , print = FALSE)
#' print(Dl)                                                  # 2-component listenv
#' print(Dl[[1]])                                             # unencoded columns
#' print(Dl[[2]])                                             # encoded sparse matrix
#'
#' # iris data can be retrieved from the Dl list in similar fashion described above
#'
#' unlink(tempf)
#'
#' }
#'

tileHot = function(readpath, rows, splits, omc = 'dgCMatrix', ...) {
                    encode = as.symbol('encode')
                  readpath = if (file.exists(readpath)) {normalizePath(readpath)
                             } else {stop('no file at source location!', call. = FALSE)}
                  delayedAssign('rn.x', 1:n.x )
                      info = match.fun(info, descend = FALSE)
                      ptot = r_bg(function(info, readpath) {
                                                         info(readpath, show.info = FALSE)[[1L]]
                               }, args = list(info, readpath), user_profile = FALSE, supervise = TRUE)
                  on.exit(M <- NULL, add = TRUE); on.exit(rez <- NULL, add = TRUE); on.exit(en <- NULL, add = TRUE)
                       evd = if (...length() > 0L) {
                                    as.list(match.call(expand.dots = FALSE)$...)}
                       dropcols = eval(evd$dropcols)
                       orn = eval(evd$orn)
                       omc = match.arg(omc, choices = c('matrix', 'dgCMatrix'))
                     prest = if (!is.null(dropcols) && isTRUE(orn)) {
                                                r_bg(function(fread, readpath, dropcols) {
                                                              fread(readpath, select = dropcols)
                              }, args = list(fread, readpath, dropcols), user_profile = FALSE, supervise = TRUE)}
                         r = splitH(readpath)
                       n.x = if('ready' %in% ptot$poll_io(500)['process']) {
                                                                             ptot$get_result()}
                                                                      else {
                                                                             ptot$wait()
                                                                             ptot$get_result()}
                       if (!is.null(dropcols)) message('- columns dropped:', sQuote(dropcols), '\n')
                         v = listenv(); en = listenv(); idx = double(); n = structure(character(), names = NULL)
                   subsets = ceiling(n.x/rows)
                     for (i in 1:subsets) {
                       if (splits > 0L) {
                            v[[i]] = do.call(splitV, list(r(rows, ...), splits))
                            for (j in seq(along=v[[i]])) {
                              if (exists('srn', as.environment(v[[i]][[j]]))) {
                                                      idx = v[[i]][[j]]$srn
                                                      v[[i]][[j]][, let('srn', NULL)]}}
                            v[[i]] = lapply(v[[i]], \(k) k[, lapply(names(.SD), \(i) paste0(i, '.', .SD[[i]]))])
                    } else if (splits == 0L) {
                                  v[[i]] = r(rows, ...)
                             if (exists('srn', as.environment(v[[i]]))) {
                                                     idx = v[[i]]$srn
                                                     v[[i]][, let(srn, NULL)]}
                                  v[[i]] = v[[i]][, lapply(names(.SD), \(i) paste0(i, '.', .SD[[i]]))]}
                            n = c(n, unique(rapply(v[[i]], f = \(i) Reduce(\(x,y) `c`(x,y), i), how = 'unlist')))
                      en[[i]] = rapply(object = v[[i]], f = oneHot, how = 'unlist', type = encode)
                       v[[i]] <- NULL
                      if (isTRUE(orn)) {
                        lapply(seq(along=en[[i]]), \(j) dimnames(en[[i]][[j]])[[1L]] <- idx)
                   }}
                   if (isTRUE(orn)) {
                           ll = unique(n); n.l = length(ll); n <- NULL
                            M = eval(M)
                           iU = unlist(en, TRUE, TRUE)
                     for (i in iU) {
                            v = as(i, omc)
            try(M[fmatch(rownames(v), rownames(M), 0L, NULL), fmatch(colnames(v), colnames(M), 0L, NULL)] <- v)}
                  if (!is.null(dropcols)) {
                         rest = if('ready' %in% prest$poll_io(500)['process']) {
                                                                                 prest$get_result()}
                                                                          else {
                                                                                 prest$wait()
                                                                                 prest$get_result()}
                  message('\n\nstructuring the output may take a while ...\n')
                                if (omc %in% 'matrix') {
                                  rez = data.table(rest, M)}
                                else {
                                  rez = listenv(rest, M)}
                                    r = idx = rest <- NULL
                                cat('\nFinished!\n\n')
                                return(rez)} else return(M)
                   } else return(en)}

test_that("matrix output", {
                  data(iris)
                     tempf = tempfile(fileext = '.csv')
                     write.table(iris,tempf , sep = ',', row.names = FALSE, quote = FALSE)
                  skip_on_os('mac')
                     C = tileHot(tempf, rows = 14, splits = 3, orn = TRUE, print = FALSE)
                     c = getEV(C, 'Species', 'matrix')
                  expect_identical(class(C)[1L], "dgCMatrix")
                  expect_identical(class(c), c("matrix", "array"))
                  expect_identical(colnames(c), c("setosa", "versicolor", "virginica"))
                  unlink(tempf)

})

#'@title Find Substring Combinations In Parallel
#'
#'@description
#'Internal function for \"fcommon\" - the parallelised version of function \"core\".
#'
#'@param chain character, length 1. A string
#'@param brows integer, length 1. Number of combinations matrix rows to be sent
#' to each worker (logical cpu). Default 1e6
#'@param tpe character, length 1. Strategy for the multiprocess plan. Choices: 'multisession',
#' 'multicore' and 'cluster'. Default 'sequential' corresponding to \code{tpe = NULL}
#'@param wo integer, length 1. Number of workers in the multiprocess plan
#'@param optMax integer, length1. Limit for size of objects sent to each worker. Default 500 MiB
#'
#'
#'@keywords internal
#'@noRd
#'


corePar = function(chain, brows = 1e6, tpe = strategy, wo = workers, optMax = maxSize) {
             'strategy' = 'workers' = 'maxSize' = NULL
            if (!is.null(tpe)) {with(future::plan(strategy = tpe, workers = wo), local = TRUE)}
            if(!is.null(optMax)) {
              ops = options(future.globals.maxSize = optMax)
                    on.exit(options(ops), add = TRUE)
            }
             sift = match.fun(sift, descend = FALSE)
                b = strsplit(as.character(chain), '')[[1L]]
               nc = nchar(chain)
                w = 2L:(nc - 1L)
                m = listenv()
              for (i in seq(along=w)) {
                  tot = comboCount(b, w[i])
                   rg = seq(1L, tot, by = brows)
    m[[i]] = future_sapply(rg, \(lower) {
             end = min(lower + brows - 1L, tot)
         comboGeneral(b
                    , m = w[i]
                    , FUN = sift
                    , lower = lower
                    , upper = end
                    , chain = chain
                    )
         }, simplify = TRUE, USE.NAMES = FALSE)

   m[[i]] = m[[i]][nchar(m[[i]]) > 1L] |> unique()
              }
          unlist(m) |> unique()
}

#' @import data.table
#' @import RVerbalExpressions
#' @import RcppAlgos
#' @import methods
#' @importFrom utils type.convert getFromNamespace
#' @importFrom fastmatch fmatch
#' @importFrom stats median setNames
#' @importFrom callr r_bg
#' @importFrom Matrix Matrix rowSums rsparsematrix
#' @importFrom listenv listenv
#' @importFrom future.apply future_sapply future_lapply
#'

    .onLoad = function(libname, pkgname) {
                         Mmatrix = getFromNamespace('Mmatrix', 'Matrix')
                   rsparsematrix = getFromNamespace('rsparsematrix', 'Matrix')
                   requireNamespace("future", quietly = TRUE)
                }

#'@title Extract A Proportional Stratified Sample From A Data Set
#'
#'@description Obtains a proportional stratified sample from any data convertible to "data.table"
#' class containing categorical variables.
#'
#'@param X any data array convertible to "data.table" class
#'@param target character length 1. The name of column considered to be the root stratum. For example, the name of the
#' 'target' categorical column in a classification training set. This argument should always have a value
#'@param stratum character of length <= \code{ncol(data) - 1}. Default, NULL. Names of additional categorical data
#'  columns which deepen the stratification
#'@param size integer length 1. Default, none. Value set by User. In this case, it is upper-bounded by the size of the
#'  thinnest stratum having more than one row. Setting \code{size} value above this bound requires sampling with
#'  replacement
#'@param thresh integer, length 1. Default, none. An automatic switch between sample size calculation formulae.
#'  Can be set when \code{size} is missing from call. It can take as value any of the stratum thicknesses shown in the
#'  output message
#'
#'  NOTE: it is recommended that \emph{both} \code{size} and \code{thresh} values are missing from call until
#'  information on stratification becomes available after first run
#'@param seed integer length 1. Seed value for output reproducibility
#'@param indx logical. Default TRUE, returns the sample row index only. FALSE, returns the sampled data
#'@param dis symbol. Default NULL. One of the density or function \link[stats]{distributions} used for
#'  generating probability vectors for probabilistic sampling
#'@param args list of arguments required by distributions as described in [stats::distributions] documentation.
#'  Default, none. \strong{NB} The list should \emph{never} include the first argument (\strong{x} or \strong{n})
#'  required in documentation, as it is collected internally from each stratum
#'
#'  NOTE: Even if \code{seed} is set, the sample row index changes if either the distribution in \code{dis} or the
#'  values in \code{args} is changed
#'@param ext logical, default FALSE. When TRUE, expands the output sampled data with the following extra columns:
#'  \strong{row} - sample rows, \strong{strat} - stratum, \strong{n} - stratum total rows (i.e. thickness)
#'  and \strong{size} - the sample size extracted from each stratum. Requires \code{indx = FALSE}
#'@param replace logical, default FALSE. When TRUE, sampling with replacement if \code{size} is present in call
#'  and exceeds the thinnest stratum with more than one row
#' @param verbose logical, default TRUE, display messages
#'
#'@details This utility is designed to find a true sample representation of the data under current stratification
#' by matching closely the proportionality of strata as long as argument \code{size} is missing from call.
#' Each distinct combination of \code{target} and \code{stratum} levels defines a stratum. For minimal
#' stratification, argument \code{target} must always have a value present in call. All one-row strata, when
#' formed, are simply appended to the compounded output.
#'
#' \code{size}. As column in the extended output, it represents the size of the sample extracted from each
#' stratum, internally derived to be proportional to stratum thickness, unbounded by the thinnest stratum with more
#' than one row. Deep stratification along with high cardinality and imbalance may severely restrict the size of the
#' compounded output which is the sum of all stratum sizes plus the number of one-row strata. The sampling occurs at
#' stratum level except for one-row strata for which \code{size = 0} is interpreted as "no sampling".
#'
#' As function argument, \code{size} is interpreted as the largest sample size without replacement that can be requested,
#' being bounded by the thinnest stratum with more than one row. The presence of \code{size} in call alters
#' the proportionality since each stratum - except one-row strata - contributes equally to the output size which is
#' the number of strata times the \code{size} value plus the number of one-row strata.
#'
#' \code{thresh}. Automatic switch that modifies stratum sample size calculation method based on the extreme stratum
#' thickness values, stratification depth and total data rows. Internally, it searches for the formula that finds
#' at least one sample size accommodating the thinnest stratum with more than one row. Messages are displayed at runtime
#' although, in most cases the condition is satisfyed at first iteration. When \code{thresh >= nrow(data)}, each stratum
#' is sampled proportional with the ratio between thinnest and thickest strata, which may lead to a relatively small size
#' output. All other \code{thresh} values compromise slightly between output size and proportionality (see Example 3).
#'
#' ## Probabilistic Sampling
#'
#' \code{dis}. The \code{prob} argument in [base::sample] cannot be used as required since the length of probability vector
#' varies with stratum thickness. Herein, stratum probability vectors are determined by the distribution specified in
#' argument \code{dis} which associates each stratum with a probability vector of thickness length. When \code{args} is
#' missing from call, \code{dis} uses the default argument values for respective distribution. An error is thrown when the
#' probability vector has insufficient number of non-zero values. See package \strong{stats}, "Distributions" documentation.
#'
#' NOTE: The random variate generators i.e. the \strong{r*} version of \link[stats]{distributions}, generate vectors of absolute
#' \emph{random deviate} values which play the role of pseudo-probabilities conformant with the requirements listed in
#' [base::sample] documentation.
#'
#'@returns A proportional or non-proportional stratified sample (depending on whether \code{size} is absent or present
#' in call), either as row index or as sampled data, compounded from random or probability samples taken from each
#' stratum. Informative messages are displayed. Existing data row names are preserved in the output case in which, the sampled
#' data output gains the column named "rn".
#'
#'@seealso \link[base]{sample}, \link[stats]{distributions}
#'
#'@keywords sampling
#'
#'@export
#'
#'@examples
#'
#' if (interactive()) {
#'
#' # 1. Row index for sampling
#'
#' data(mtcars)
#' rowID = stratify(mtcars
#'                , target = 'cyl'
#'                , stratum = c('vs', 'am')
#'                , seed = 314)                                  # display information
#' print(rowID)                                                  # integer
#'
#' # 2. Sampled data with extra-columns
#'
#' smp = stratify(mtcars
#'             , 'cyl'
#'             , c('vs', 'am')
#'             , seed = 314
#'             , indx = FALSE
#'             , ext = TRUE)                                     # extra columns
#' print(smp)
#' identical(rowID, smp$row)                                     # TRUE
#'
#' # 3. Impact of "thresh" value on output size
#'
#' sl = list()
#' thresholds = c(2, 4, 12, 32)                                  # stratum thicknesses
#'
#' for (t in seq(along=thresholds)) {
#'                   sl[[t]] = stratify(mtcars
#'                                   , 'cyl'
#'                                   , c('am', 'vs')
#'                                   , thresh = thresholds[t]
#'                                   , seed = 314
#'                                   , indx = FALSE, ext = TRUE)
#'                 }
#' names(sl) = quote(thresholds)
#' print(sl)                                                     # stratified samples
#'                                                               # of various sizes
#'
#' # 4. Probabilistic sampling
#'
#' rowIDn = stratify(mtcars
#'              , 'cyl'
#'              , c('vs', 'am')
#'              , seed = 314
#'              , dis = pnorm                                    # Normal distribution
#'              , args = c(mean = 1, sd = 3))                    # no first argument!
#' rowIDb = stratify(mtcars
#'              , 'cyl'
#'              , c('vs', 'am')
#'              , seed = 314                                     # same seed
#'              , dis = pbeta                                    # Beta distribution
#'              , args = c(shape1 = 1, shape2 = 3))              # no first argument!
#'
#' # Same seed but changing the distribution changes the sample row index
#' identical(rowIDn, rowIDb)                                     # FALSE
#'
#'}
#'

stratify = function(X, target, stratum = NULL, size, thresh, seed = NULL, indx = TRUE
                  , dis = NULL, args = list(), ext = FALSE, replace = FALSE, verbose = TRUE) {
                   n = strat = Mx = j = NULL
                   stopifnot('\n \"target\" should not be part of \"stratum\"!\n'= isTRUE(!target %in% stratum))
              if (!is.data.table(X)) data = as.data.table(X, keep.rownames = TRUE) else data = X; X <- NULL
                 strata = if (is.null(stratum)) quote(target) else quote(c(target, stratum))
                      s = match.fun(s, descend = FALSE)
         checkCondition = match.fun(checkCondition, descend = FALSE)
                 sample = match.fun(sample, descend = FALSE)
                 delayedAssign('Mx', max(dtt$n, na.rm = TRUE)); delayedAssign('m', min(dtt[n > 1L]$n, na.rm = TRUE))
                 delayedAssign('M', sum(dtt$n, na.rm = TRUE)); delayedAssign('med', ceiling(median(dtt[n > 1L]$n, na.rm = TRUE)))
                data[, let(c('row', 'strat', 'n'), list(.I, .GRP, .N)), by = eval(strata), env = list(strata = strata)]
                    n.d = names(data)
              if (missing(size)) {
                    dtt = unique(data, by = 'strat')
                 thresh = if (missing(thresh)) 1L else thresh
                    dtv = dtt[, list(size = fifelse(n == 1L, 0L, s(n, m, med, Mx, M, thresh)), strat = strat, n = n)]
                     se = dtv$size; ss = unique(se); ne = dtv$n
              condition = any(se > ne) || all(ss > m)
                 if (isTRUE(condition)) {
                   message('\n\nfound inadequate sample size, switching ...\n')
                    dtv <- NULL
                    dtt = do.call(checkCondition, list(m, med, M, dtt))
             } else dtt = dtv
                     ss = unique(dtt$size)
               if (verbose) cat('\n- strata sample sizes are:', sort(ss),
                       '\n- median and thickest strata have', med,'and', Mx,'rows each\n')
                   data = dtt[data, on = 'strat', allow.cartesian = TRUE][, let(grep('i\\.', names(.SD)), NULL)]
                   setcolorder(data, n.d)
               if (verbose) cat('- maximum sample size without replacement is', m, 'under current stratification\n\n')
              } else size
                     ss = m = med = M = Mx = dtv <- NULL
             if(!is.null(seed)) set.seed(seed)
             if (!is.null(dis)) args = c(quote(row), args)
                     id = data[, if(any(n == 1L)) row else j, by = strat, env = list(j = CALLs)]$id
            return(eval(retz))
   }

test_that("data sampling pays", {
                                  data(mtcars)
                             ID = stratify(mtcars
                                         , 'cyl'
                                         , c('vs', 'am')
                                         , seed = 314)
                             rowID = stratify(mtcars
                                            , 'cyl'
                                            , c('vs', 'am')
                                            , seed = 314
                                            , indx = FALSE
                                            , ext = TRUE)$row
                             expect_equal(ID, rowID)
})

#' @title Functional Utilities For Data Processing
#' @name akin-package
#'
#' @description
#' The intent of this package is utilitarian which explains the relatively large number of informative messages
#' displayed by most functions. Designed for large and very large data files, the package employs data.table,
#' list environments, sparse matrices, iterators, background and parallel processing in few places. Nevertheless,
#' Users could consider thread optimization or memoisation, as these techniques were outside the scope.
#'
#' The package covers several areas of data processing: subset reading and writing of large data files, data tiling
#' (horizontal and vertical splitting) - suited for data conversion operations with local as well as global hold,
#' such as one-hot encoding - stratified, proportional, random or probabilistic data sampling, data normalization and
#' thresholding, substring location inside strings e.g. peptides inside protein chains, common substrings identification
#' and location-tabulation of amino acids, modifications or their associated monoisotopic masses inside modified peptides
#' covering various representations of protein mass spectrometry data with no pretense of exhaustiveness.
#'
#' Comments and suggestions regarding tricky situations are provided however brief. Examples should be run individually
#' in R console. When possible, suppression of messages should be avoided.
#'
#' @keywords internal
#'
 "_PACKAGE"
 NULL

#' @title Extract Encoded Variables From Encoded Split Or Tiled Data
#'
#' @description Extracts a single encoded variable from a \link[base]{list} or \link[listenv]{listenv}
#' of encoded matrices containing multiple encoded variables
#'
#' @param en a (named) list, or \link[listenv]{listenv} of matrices or a single matrix, all containing multiple
#'   encoded variables. See [oneHot] decoder for lists and matrices containing single encoded variables
#' @param name character, length 1. Column name as found in source data
#' @param ... default, empty. Used to convert the class of extracted matrix to 'dgCMatrix' or 'matrix'
#'
#' @details This function includes code from package "Matrix.utils" v 0.9.8, published under GPL-3
#' license, currently removed from CRAN. With thanks to the package Author!
#'
#' NOTE 1: If \code{name} is a source data column name that appears inside other column names, the extracted matrix
#' will combine all encoded matrices having this name inside their column names. Although the extracted matrix is a
#' proper matrix of encodings, it no longer represents \emph{a single} encoded data column. As result, upon decoding,
#' the [oneHot] decoder will report ambiguous decoding.
#'
#' NOTE 2: a warning reading either \emph{"single-column encoded matrix for ..."} or \emph{"number of columns of
#' result is not a multiple of vector length (arg 2) ..."} may appear when extracting an encoded categorical variable
#' from a list of encoded matrices. Most likely, this happens with low cardinality encoded variables. The warning
#' signals that most encoded matrices associated with respective variable contain subsets of only one category (level)
#' when, ideally, most of these matrices should contain a mixture of two or more categories or levels; thus, allowing
#' matrix row-binding by category's label. One or more of the following suggestions will solve the issue: a) shuffle
#' the data before encoding, b) increase the number of \code{rows} in data chunks when encoding, c) if memory allows,
#' opt for [tileHot] encoding single matrix output, as shown in Example 2.1, solution c.
#'
#' @returns A dense or sparse matrix of single encoded variable which can be decoded with the [oneHot] decoder.
#'
#' @seealso \link{oneHot}, \link{tileHot}
#'
#' @keywords extractor
#'
#' @export
#'
#' @examples
#'
#' if (interactive()) {
#'
#' # 1. mtcars data have all columns type "double"
#'
#' data(mtcars)
#' a = lapply(mtcars, oneHot, encode)                       # encode mtcars data
#' print(a)                                                 # list of sparse matrices
#' b = getEV(a, 'cyl')                                      # extract encoded "cyl" column
#' print(b)                                                 # a 32x3 sparse matrix
#' c = oneHot(b, decode)                                    # revert
#' identical(mtcars$cyl, c)                                 # FALSE. 'mtcars$cyl' is type "double"
#' isTRUE(all.equal(mtcars$cyl, c))                         # TRUE
#'
#' # 2. Warnings associated with low cardinality categorical variable
#'
#' # See tileHot() Examples for full decoding of a dataset
#'
#' # 2.1 Make 'csv' file
#' data(iris)                                               # low cardinality "Species"
#' tempf = tempfile(fileext = '.csv')
#' write.table(iris, tempf , sep = ',', row.names = FALSE, quote = FALSE)
#'
#' A = tileHot(readpath = tempf, rows = 14, splits = 3)     # encoded tiles list
#' print(A[[11]][[5]])                                      # e.g. one-column matrix
#' a = getEV(A, 'Species')                                  # warning
#' colSums(a)                                               # incorrect!
#'
#' # solution b
#' B = tileHot(readpath = tempf, rows = 60, splits = 3)     # increase number of rows
#' b = getEV(B, 'Species')                                  # still warning
#' colSums(b)                                               # incorrect!
#'
#' # Solution b) could work in combination with solution a)
#'
#' # solution c
#' C = tileHot(tempf, rows = 14, splits = 3, orn = TRUE)    # encoded matrix
#' c = getEV(C, 'Species')                                  # no warning
#' colSums(c)                                               # correct!
#'
#' unlink(tempf)
#'
#' # 2.2 Shuffled 'csv' file
#' tempf = tempfile(fileext = '.csv')
#' iris22 = iris[{ set.seed(327); sample.int(150) },]      # shuffled iris data
#' write.table(iris22, tempf , sep = ',', row.names = FALSE, quote = FALSE)
#'
#' A = tileHot(readpath = tempf, rows = 14, splits = 3)    # same as above
#'
#' #solution a
#' a = getEV(A, 'Species')                                 # no warning
#' colSums(a)                                              # correct!
#'
#' unlink(tempf)
#'
#' }
#'
#'

getEV = function(en, name, ...) {
                    on.exit(x <- NULL, add = TRUE); on.exit(r <- NULL, add = TRUE); on.exit(rr <- NULL, add = TRUE)
                   if(is.data.table(en) || is.data.frame(en)) {stop('data should be matrix or list only!', call. = FALSE)
                  } else if (is.matrix(en) || inherits(en, 'dgCMatrix')) {
                      nm = grepl(name, colnames(en))
                       x = if (...length() == 0L) en[, nm] else as(en[, nm], ...)
            if (isTRUE(sum(nm) > 1L)) {
            colnames(x) <- gsub(paste0(name, '\\.', ''), '', colnames(x))}}
                  else if (inherits(en, 'listenv') || is.list(en)) {
                    uen = unlist(en)
                    scm = sapply(seq(along=uen), \(i) dim(uen[[i]])[[2L]] == 1L)
                     tt = grepl(name, sapply(uen[which(scm == TRUE)], colnames)); uen <- NULL
                       if (isTRUE(sum(tt) > 0L)) warning('single-column encoded matrix for ', name, '!', call. = FALSE)
                       if (!is.null(names(en))) {
                          nm = grepl(name, names(en))
                           x = rBind.fill(en[nm], ...)}
                       else {
                           r = rapply(as.list(en), f = getEV, how = 'unlist', name = name)
                          rr = r[lengths(r) > 0L]
                           x = if (...length() == 0L) rBind.fill(rr) else as(rBind.fill(rr), ...)}}
                      return(x)
                  }

#'@title Defunct Functions In Package akin
#'
#'@description The following functions are defunct. Where available, alternative functions are suggested
#' at \code{help("akin-defunct")}.
#'
#' @section \code{common}: this function is defunct. Use [fcommon] instead.
#' @keywords internal
#' @name akin-defunct
#'
NULL

#' @title Scaling And Thresholding Of Numeric Variables
#'
#' @description Implements classical methods for data scaling: range, z-score normalization, location and location-scale
#' normalization, as well as data thesholding through the simplest form of ReLU rectifier. Missing values are removed in
#' all cases.
#'
#' @param x numeric vector, length > 1. Variable to be scaled or filtered
#' @param how symbol. Choices are \strong{range}, \strong{stdev} or \strong{relu}
#' @param filter character, length 1. Default NULL. Choices "positive", "negative". Requires \code{how = relu}
#' @param ... list reserved for User input of paired values, statistic or otherwise. The list uses individual
#' ellipsis arguments therefore, order of values needs be respected at all times e.g. when \code{how = range}  min value
#' \emph{first}, max value \emph{second}. When \code{how = stdev}, mean value \emph{first}, standard deviation \emph{second}.
#'
#' @details
#' Normalization (scaling) can be applied locally on subsets of \code{x} when User inputs the values in the \code{...} list.
#' Otherwise, the scaling is global i.e. it is applied to \code{x} as a whole. No assumptions regarding the underlying distribution
#' of \code{x} are made.
#'
#' \code{how != relu}. When \code{...} is empty, the function uses the sample statistics of \code{x} e.g. the mean, range
#' or standard deviation. Otherwise, it uses values inputted by User case in which, \emph{location-scale} normalization
#' requires \code{how = stdev} and the \code{...} list filled as follows: the \code{min(x)} \emph{or} \code{max(x)}, or any other
#' value first and \code{sd(x) > 0} or any other positive value second. In particular, \emph{location} normalization works similarly
#' but with the second value = 1. Other location types, e.g. x/max(x), are obtainable.
#'
#' NOTE: when \code{...} is populated with custom values, all other arguments should be present and named (see Examples).
#'
#' \code{how = relu}. This option acts as numeric thresholding locally as well as globally. It stands for \strong{re}ctified
#' \strong{l}inear \strong{u}nit and involves no statistics. It applies to numeric types that have ordering property (double,
#' integer). On return, all \code{x} attributes are dropped. When \code{filter = 'positive'}, all negative values are set to zero
#' while positive values remain unchanged. Alternatively, when \code{filter = 'negative'}, all negative values remain unchanged
#' while all other values are set to zero. The "negative" option was added for symmetry.
#'
#' @returns Numeric. When missing \code{...} and \code{how != relu}, scaled values using \code{x} own sample statistics.
#' Otherwise, scaling is based on values inputted by User. When \code{how = relu}, \code{x >= 0} or \code{x <= 0}, depending
#' on \code{filter} setting.
#'
#' @references \href{https://en.wikipedia.org/wiki/Ancillary_statistic}{Ancillary
#'   Statistic} for location and location-scale distributions
#'
#'@keywords thresholding
#'
#' @export
#'
#' @examples
#'
#' if (interactive()) {
#'
#' # 1. ReLU thresholding
#'
#' x = { set.seed(223); sort(runif(10, -3, 3)) }
#' y = score(x, relu, 'positive'); y
#' z = score(x, relu, 'negative'); z
#'
#' # 1.1 ReLU Plot
#' olp = par(no.readonly = TRUE)
#' par(list(mar = c(1,1,1,1), mgp = c(0,0,0), tcl = -0.01, pty = 's'))
#' plot(x, y, type = 'l', col = 'steelblue', lwd = 2 ,
#'        xlim = c(min(x), max(x)), ylim = c(min(x), max(x))
#'      , ylab = expression(ReLU(x)), xaxs = 'i', yaxs = 'i', axes = FALSE, cex.lab = 0.7)
#' axis(1, pos = 0, cex.axis = 0.6) ; axis(2, pos = 0, cex.axis = 0.6)
#' points(x, z, type = 'l', col = 'orangered', lwd = 2)
#' legend('topleft', legend = c('positive', 'negative'),
#'        col = c('steelblue', 'orangered'), pch = 'l', lwd = 2, cex = 0.6, bty = 'n')
#' par(olp)
#'
#' # 2. Location and location-scale
#'
#' # 2.1 Location (e.g. "x - max(x)")
#' x = 1:10
#' M = max(x)
#' std = 1
#' a = score(x, stdev, NULL, M, std); a
#'
#' # 2.2 Location (e.g. "x/max(x)")
#' m = 0                                                            # the mean
#' M = max(x)                                                       # or any value
#' b = score(x, range, NULL, m, M); b
#'
#' # 2.3 Location-scale (e.g. "(x - max(x))/sd(x)")
#' M = max(x)                                                       # or any value
#' std = sd(x)                                                      # or any value > 0
#' c = score(x, stdev, NULL, M, std); c
#'
#' # m, M and std above can be replaced with any values decided by User
#'
#' # 3. Classical normalization
#'
#' # 3.1 Range
#' d = score(x, range); d
#'
#' # 3.2 z-score
#' e = score(x, stdev); e
#'
#' # 4. Local vs. global z-score normalization
#'
#' data(mtcars)
#'   x = mtcars$wt
#'   m = mean(x)
#' std = sd(x)
#'
#' ll = split(x, f = as.factor(mtcars$cyl))        # partitioned x
#'
#' # 4.1 Local scaling
#' aa = lapply(ll, score, stdev, NULL, m, std)     # filled ... list
#' na = unlist(aa, FALSE, FALSE)
#'
#' # 4.2 Global scaling
#' nb = score(x, stdev)
#'
#' # 4.3 Local as well as global hold
#' identical(sort(na), sort(nb))                   # TRUE
#'
#' }
#'

score = function(x, how, filter = NULL, ...) {
                       r = NULL
                     how = substitute(how)
                     how = match.arg(as.character(how), choices = c('range', 'stdev', 'relu'))
               stopifnot('x must be numeric vector of length > 1' = is.numeric(x) && length(x) > 1L)
               if(anyNA(x)) {
                            warning('\nall missing values were removed!', call. = FALSE)
                     x = x[!is.na(x)]}
                   n.x = length(x)
                     M = call('max', x, na.rm = TRUE); m = call('min', x, na.rm = TRUE)
                  mean = call('mean', x, na.rm = TRUE); sd = call('sd', x, na.rm = TRUE)
              switch(how
                 , range = {
                   if (...length() == 0L) {
                         M = eval(M)
                         m = eval(m)
                   stopifnot('zero range found!'= max(m, M) > min(m, M))
                   } else {
                        m = ..1
                        M = ..2
                    if (any(!is.numeric(m), !is.numeric(M))) stop('quantities must be numeric!', call. = FALSE)
                         }
                       if(!is.null(m) && !is.null(M)) r = max(m, M) - min(m, M)
                       x[1:n.x] <- (x - min(m, M))/r
                       }
                , stdev = {
                   if (...length() == 0L) {
                     mean = eval(mean)
                       sd = eval(sd)
                  stopifnot('zero standard deviation found!' = sd(x) > 0)
                  } else {
                     mean = ..1
                       sd = ..2
                    if (any(!is.numeric(mean), !is.numeric(sd))) stop('quantities must be numeric!', call. = FALSE)
                       }
                     x[1:n.x] <- (x - mean)/sd
                   }
                , relu = {
               filter = match.arg(filter, choices = c('positive', 'negative'))
                 if (is.null(filter)) stop('argument \"filter\" should be either \"positive\" or \"negative\"!', call. = FALSE)
                    x <- if (filter %in% 'positive') pmax.int(0, x, na.rm = TRUE) else pmin.int(0, x, na.rm = TRUE)
                }
                , character(0))
              return(x)
            }

test_that("normalization-threshlding", {
                                  x = { set.seed(223); sort(runif(10, -3, 3)) }
                                  y = score(x, relu, 'positive')
                             expect_identical(isTRUE(all(y >= 0)), TRUE)
                                  z = 1:9
                                 zz = z/max(z)
                                  A = score(z, range, NULL, 0, 9)
                             expect_equal(A, zz)
})

#' @title Tile And Write Tiled Data To Disk
#'
#' @description Splits long and wide data files in lists of disjoint tiles for further processing.
#'
#' @param readpath character length 1. Full path to the source file
#' @param writepath character length 1. Full path to the destination file
#' @param rows integer length 1. Number of rows in each subset. Internally, it determines the total
#'    number of subsets before the vertical split
#' @param splits integer, length 1. Number of vertical data splits in each above subset. See [splitV]
#' @param ... extra arguments to [splitH] e.g. \code{dropcols} for columns dropped from source data
#'
#' @details Facilitates local operations on small size tiles by partitioning the data horizontally and vertically.
#' The list of tiles can be written to disk as "rds" file when a \code{writepath} destination is given. The written
#' data can then be read entirely or in subsets (see Example 2).
#'
#' NOTE: This utility uses background processing. Check "Security Considerations" in \strong{callr} package documentation.
#'
#' @returns A \link[listenv]{listenv} of "data.table" class tiles. When \code{writepath} is given, it produces
#' a "rds" file containing data tiles.
#'
#' @seealso [splitH], [splitV], [tileHot], \link[base]{readRDS}
#'
#' @keywords splitting
#'
#' @export
#'
#' @examples
#'
#' if (interactive()) {
#'
#' # Make a 'csv' file
#'
#' data(iris)
#' tmpf = tempfile(fileext = '.csv')
#' write.table(iris, tmpf , sep = ',', row.names = FALSE, quote = FALSE)
#'
#' # 1. Tile data
#'
#' a = tileData(tmpf, rows = 10, splits = 3)     # 10x2 and 10x1 tiles
#' class(a)                                      # listenv, environment
#' str(a)                                        # nested list
#'
#' tmpf1 = tempfile(fileext = '.rds')            # new location
#'
#' # 2. Write tiled data
#'
#' tileData(tmpf, tmpf1, rows = 10, splits = 3)
#' a = readRDS(tmpf1)[[1]]                  # partial read from new location
#' print(a)                                     # list component
#'
#' unlink(tmpf)
#' unlink(tmpf1)
#' }
#'
#'

tileData = function(readpath, writepath = NULL, rows, splits, ...) {
              on.exit(w <- NULL, add = TRUE)
                v = listenv()
         readpath = if (file.exists(readpath)) {normalizePath(readpath)
                    } else {stop('no file at source location!', call. = FALSE)}
             info = match.fun(info, descend = FALSE)
             dims = r_bg(function(info, readpath) {
                                          info(readpath, show.info = FALSE)
                            }, args = list(info, readpath), user_profile = FALSE, supervise = TRUE)
                r = splitH(readpath)
              if('ready' %in% dims$poll_io(100)['process']) {
                           tot = dims$get_result()[[1L]]
                          nmes = dims$get_result()[[2L]]} else {
                                                                 dims$wait()
                                                                 tot = dims$get_result()[[1L]]
                                                                nmes = dims$get_result()[[2L]]}
          subsets = ceiling(tot/rows)
           for (i in 1:subsets) v[[i]] = splitV(r(rows, ...), splits)
                w = lapply(v, as.list)
           if (!is.null(writepath) && is.character(writepath) && length(writepath) == 1L) {
                     invisible(saveRDS(w, writepath))
             message('\nfile saved to', path.expand(writepath), '\n')
           } else return(w)
     }

#' @title Conditions For Function \"fcommon\"
#'
#' @keywords internal
#' @noRd
#'

    cond = substitute(expr = isTRUE(grep(y, chain, useBytes = TRUE, ...) > 0L))

#' @title Substring Validation Used In Function \"fcommon\"
#' @description Checks combinations for appartenance to source chain
#'
#' @keywords internal
#' @noRd
#'

 sift = function(x, chain, ...) {
              y = paste0(x, collapse = '')
              if (eval(cond)) y else NULL
}

#' @title Object For oneHot Function, decode
#' @description Logical expression
#'
#' @keywords internal
#' @noRd
#'

 tof = substitute(expr = isTRUE(
         any(sapply(n.x, \(n) is.double(type.convert(n, as.is = TRUE, numerals = 'no.loss')))))
         )


#' @title The j-side Of \"data.table\" Sample Calculation
#'
#' @keywords internal
#' @noRd
#'

 CALLs = as.call(
   c(quote(list), id = substitute(sample(row, if(length(size) > 0L) size = unique(size) else size = size, replace = replace, prob = if (!is.null(eval(dis))) abs(do.call(dis, args)))
                , env = list(size = quote(size), replace = quote(replace), args = quote(args), dis = quote(dis))))
   )

#' @title Return Value For Sample Calculation
#'
#' @description The \"return()\" Code Part Of Function \"stratify\" Function
#'
#' @keywords internal
#' @noRd
#'

 retz = substitute(
         expr = if(isTRUE(indx)) id else if(isTRUE(ext)) data[id][] else {
                   if (missing(size)) data[id][, let(c('size', 'row', 'strat', 'n'), NULL)][]
                   else data[id][, let(c('row', 'strat', 'n'), NULL)][]
                })


#'@title Switch Formulas For Sample Size Calculation
#'
#'@keywords internal
#'@noRd
#'

 s = function(n, m, med, Mx, M, thresh) {
            if(is.numeric(thresh) && is.double(thresh)) ceiling(thresh) else thresh
            if(thresh < M) {
               if(thresh <= m) {ceiling(sqrt(n)*log(n))
               } else if (m < thresh & thresh <= med) {ceiling(sqrt(n)*log(n, base = 4))
               } else ceiling(n*(med/Mx))
            } else ceiling(n*(m/Mx))
}

#' @title Checks If Calculated Sample Size Acommodates The Thinnest Stratum
#'
#' @keywords internal
#' @noRd
#'

checkCondition = function(m, med, M, dtt) {
                        n = strat = Mx = NULL
                        on.exit(return(dtt), add = TRUE)
                        for (t in c(m, med, M)) {
                                dtt = dtt[, list(size = fifelse(n == 1L, 0L, s(n, m, med, Mx, M, thresh = t)), strat = strat, n = n)]
                                 ss = unique(dtt$size)
                             if (any(0L < ss && ss <= m)) {
                              message('\nfound valid formula for ', t, ' rows\n')
                              break
                          } else message('\nSwitching again ...\n')}
                     }

#' @title Internal Function For Function \"splitH\"
#'
#' @keywords internal
#' @noRd
#'

 rename = function(nmes) call('lapply', X = setNames(nm = nmes), FUN = as.name)

#' @title Other Objects for \"splitH\"
#'
#' @keywords internal
#' @noRd
#'

  srn = substitute(expr = if (is.infinite(rows)) NULL else if (orn && is.finite(rows)) .I + i*rows)
 jump = substitute(expr = if (is.finite(rows)) i*rows + 1L else 0L)


#' @title Conversion Utility
#'
#' @description
#' Internal function used in decoding a matrix of encoded data
#'
#' @param x,y sparse or dense vectors, the result of encoding
#'
#' @returns decoded vector
#'
#' @keywords internal
#' @noRd
#'
#'

  g = function(x, y) {
                       z = fifelse(x == 1L, y, NA)
   return(type.convert(z[!is.na(z)], as.is = FALSE, numerals = 'no.loss', tryLogical = FALSE))
          }


  #' @title Display Brief Information On A Data File
  #'
  #' @param readpath character, length 1. Full path to data file
  #' @param show.info logical. Default TRUE: shows information on data file
  #'
  #' @references The code was inspired by
  #' \href{https://github.com/Rdatatable/data.table/issues/7169}{Issue# 7169}
  #'
  #' @details Provides data [base::dim] and [base::names] without loading the
  #' entire file into memory.
  #'
  #' @returns Prints info and invisibly returns the number of rows, and data file
  #' column names for internal use.
  #'
  #'@keywords internal
  #' @noRd
  #'


  info = function(readpath, show.info = TRUE) {
         tot = as.integer(strsplit(system2("wc", c("-l", readpath), stdout = TRUE, wait = FALSE), " ", fixed = TRUE)[[1L]][1L]) - 1L
        nmes = strsplit(system2("head", c("-n1", readpath), stdout=TRUE, wait = FALSE), ",", fixed=TRUE)[[1L]]
      nonmes = which(nmes %in% ""); l.no = length(nonmes)
     dupnmes = which(duplicated(nmes)); l.dup = length(dupnmes)
        if (show.info) {
          if (isTRUE(any(l.no || l.dup ))) {
          stop('Source contains ', l.no,'\\', l.dup, ' empty\\duplicated column names!!\n', call. = FALSE)}
      message("\nFound file containing: ", tot, " rows and ", length(nmes), " columns named\n\n", sQuote(nmes), "\n\n")
    }
    return(invisible(list(tot, nmes)))
  }

#'@title Get Unexported \"Mmatrix\" From Package \"Matrix\"
#'
#'@keywords internal
#'@noRd
#'

M = substitute(expr = {
         Mmatrix = utils::getFromNamespace('Mmatrix', 'Matrix')
        as(.External(Mmatrix
                   , data = 0L
                   , nrow = n.x
                   , ncol = n.l
                   , byrow = FALSE
                   , dimnames = list(rn.x, ll)
                   , mnrow = FALSE
                   , mncol = FALSE), omc)
})

#'
#' @title Row-bind Sparse Matrices
#'
#' @description [base::rbind]s a list of [Matrix::Matrix] or [base::matrix]
#' objects, filling in missing columns.
#'
#' NOTE 1: This code was taken as is from package 'Matrix.utils' v 0.9.8 published
#' under GPL-3 license and removed from CRAN. With thanks to the package Author!
#' NOTE 2: In this copy, [base::match] function was replaced with [fastmatch::fmatch]
#'
#' Extract from function's documentation:
#'
#' @param x, ...	objects to combine. If the first argument is a list and ... is
#'   unpopulated, the objects in that list will be combined.
#' @param fill value with which to fill unmatched columns
#' @param out.class the class of the output object. Defaults to the class of x.
#'   Note that some output classes are not possible due to R coercion
#'   capabilities, such as converting a character matrix to a Matrix.
#'
#' @keywords internal
#' @noRd
#'

 rBind.fill = function (x, ..., fill = NULL, out.class = class(rbind(x, x))[1]) {
      if (is.list(x) && !is.data.frame(x) && missing(...)) {
      Reduce(function(x, y) rBind.fill.internal(x, y, fill, out.class), x)
                     }
         else {
              Reduce(function(x, y) rBind.fill.internal(x, y, fill,  out.class), list(x, ...))
      }
}


 rBind.fill.internal<-function(x,y,fill,out.class) {
  out.class<-force(out.class)
  fillMissing<-is.null(fill)
  if(fillMissing)
    fill<-if(is(x,'Matrix')) 0 else NA
  if (is.null(nrow(x)))
    x<-matrix(x,nrow=1,dimnames=list(NULL,names(x)))
  if (is.null(nrow(y)))
    y<-matrix(y,nrow=1,dimnames=list(NULL,names(y)))

  nullNames<-FALSE
  if(is.null(colnames(x)))
    colnames(x)<-make.names(colnames(y)[1:ncol(x)],unique = TRUE)
  if(is.null(colnames(y)))
    colnames(y)<-make.names(colnames(x)[1:ncol(y)],unique = TRUE)
  if(is.null(colnames(x)))
  {
    nullNames<-TRUE
    colnames(x)<-1:ncol(x)
    colnames(y)<-1:ncol(y)
  }
  ymiss<-colnames(x)[which(is.na(fmatch(colnames(x),colnames(y))))]
  ybind<-rsparsematrix(nrow=nrow(y),ncol=length(ymiss),0)
  colnames(ybind)<-ymiss
  if(!fillMissing)
    ybind[seq_along(ybind)]<-fill
  xmiss<-colnames(y)[which(is.na(fmatch(colnames(y),colnames(x))))]
  xbind<-rsparsematrix(nrow=nrow(x),ncol=length(xmiss),0)
  colnames(xbind)<-xmiss
  if(!fillMissing)
    xbind[seq_along(xbind)]<-fill
  if (ncol(xbind>0))
    x<-cbind2(x,xbind)
  if(ncol(ybind)>0)
    y<-cbind2(y,ybind)
  y<-as(y,out.class)
  x<-as(x,out.class)
  result<-rbind2(x,y[,order(fmatch(colnames(y),colnames(x)))])
  if(nullNames)
    colnames(result)<-NULL
  return(result)
}

#' @title Regular Expressions And Objects For \"locateMod\" function
#' @keywords internal
#' @noRd
#'

.ARGS = quote(c('-', '+', ',', ';', ':', '=', '.', '[:digit:]', '[:alpha:]', '\\w+', ' '))
.plac = quote(rx() %>% rx_either_of(brackets) %>% rx_one_or_more())
  .uu = quote(paste0(wrap, brackets[which(brackets %in% awrap)], collapse = ''))
  .uv = quote(rx() %>% rx_either_of(awrap, wrap))
   .u = quote(rx() %>% rx_any_of(paste0(awrap, abracket, collapse = '')) %>% rx_anything_but(uu) %>% rx_any_of(paste0(inbracket, wrap, collapse = '')) %>% rx_one_or_more())
   .w = quote(rx() %>% rx_either_of(setdiff(ARGS, except)))


#' @title Filter Conditions For function \"fcommon\"
#'
#' @keywords internal
#' @noRd
#'

 findSeqUp = substitute(expr = union(ii[which(shift(ii, -1L) == ii + 1L)], ii[which(shift(ii, 1L) == ii - 1L)]))
 findSeqDn = substitute(expr = union(jj[which(shift(jj, -1L) == jj + 1L)], jj[which(shift(jj, 1L) == jj - 1L)]))

#' @title Split Data Vertically Into Subsets
#'
#' @description Splits data into unequal and disjoint groups of columns (i.e. vertical splits)
#'
#' @param data a "data.table" class data frame or convertible to "data.table" class
#' @param splits integer, length 1, of value <= \code{ncol(data)}. The number of disjoint
#'   selections (i.e. vertical splits). When \code{splits = 0}, there are no splits
#'
#' @details The smaller the \code{splits} value, the wider the column groups. Column order from
#' source data is not preserved
#'
#' @returns An exhaustive \link[listenv]{listenv} of disjoint column groups from original data
#'
#' @keywords splitting
#'
#' @export
#'
#'@examples
#'
#' if (interactive()) {
#'
#' # 1. Split iris data vertically
#'
#' data(iris)
#' a = splitV(iris, splits = 3)     # split data in 3 column groups
#' class(a)                         # listenv, environment
#' print(as.list(a))                # list
#'
#'}


splitV = function(data, splits) {
                dt = if (!is.data.table(data)) as.data.table(data) else data
              l.dt = length(dt); n.dt = names(dt)
            rename = match.fun(rename, descend = FALSE)
               stopifnot('splits value is larger than total number of columns' = splits <= l.dt)
                 y = data.table(x = n.dt, s = if (splits > 0) 1:l.dt %% splits + 1L else 1L)[order(s)]
                ll = listenv()
            for (i in unique(y$s)) {
              nmes = y[s == i]$x
                 j = as.call(c(quote(list), eval(rename(nmes), list(nmes = I(nmes)))))
           ll[[i]] = dt[, j, env = list(j = j)]}
           return(ll)
 }

#' @title Read Or Write Subsets Of Data Files From Or To Disk
#'
#' @description Reads or writes data files from/to disk in disjoint subsets. This is a two-stage function (see Examples).
#'
#'
#' @param readpath character length 1. Full path to the source file
#' @param writepath character length 1. Full path to the destination file
#'
#' @details
#' Above arguments apply to Stage 1 only. The arguments for Stage 2 function, which is the output of Stage 1,
#' are the following:
#'
#' \code{rows} integer, length 1. Number of rows per subset. When rows = Inf, the data can be either
#'     copied \emph{as is} or moved to a new location
#'
#' \code{seq} logical, default TRUE: read discrete subsets. Otherwise, progressively appended subsets
#'     from first to current
#'
#' \code{dropcols} character of \code{length < ncol(data)}. Columns to drop. Works only when \code{rows} is
#'     finite. Replaces argument \code{select} from [data.table::fread]
#'
#' \code{how} symbol. Works only when \code{rows = Inf} and \code{writepath} location is given.
#'     Options: \code{how = scp}, data file is copied \emph{as is} to \code{writepath} location;
#'     \code{how = mv}, data file is moved to \code{writepath} location
#'
#' \code{print} logical, default TRUE, each subset written to disk is shown in console. Setting print to
#'     FALSE could increase writing speed
#'
#' \code{orn} logical, default FALSE. When TRUE, the \strong{o}riginal data \strong{r}ow \strong{n}umbers
#'     are shown in each subset
#'
#'
#' The main purpose of this utility is to bring manageable subsets from very large data into the working
#' environment for further processing when \code{writepath = NULL}. When \code{orn = TRUE}, each subset
#' receives a new column named "srn" showing \strong{s}ource data \strong{r}ow \strong{n}umbers. This column
#' is absent from subsets written to disk regardless of \code{orn} value. The source data file can be any
#' type of file readable by [data.table::fread].
#'
#' At the first stage:
#'
#'   * the utility retrieves information about source data without loading them into memory and also provides
#' the new function which, in the second stage:
#'
#'   * reads source data in successive disjoint subsets (\code{rows < Inf}) and brings them into the work
#' environment (writepath = NULL), \strong{or}
#'
#'   * writes subsets to \code{writepath} location appending them automatically to the destination file.
#' During writing, if (\code{print = TRUE}) the displayed subsets are just printouts (class "NULL"). When
#' \code{writepath = NULL}, displayed subsets are objects.
#'
#' There is a functional difference between \code{rows = Inf} and \code{rows = nrow(data)}:
#'
#'   * when \code{rows = Inf}, the size of source data is irrelevant. They can be either copied (\code{how = scp})
#' or moved (\code{how = mv}) to \code{writepath} destination without being loaded into memory.
#'
#'   * when \code{rows} has finite value, the size of source data is relevant and data columns can be dropped.
#'
#' @returns At stage 1, displayed information and a function (a closure). At stage 2, a "data.table" class subset
#' of data or a printout of said subset when written on disk.
#'
#' @references Part of internal code for Stage 1 was inspired by
#' \href{https://github.com/Rdatatable/data.table/issues/7169}{data.table Issue# 7169}
#'
#' @seealso Linux commands \strong{scp} and \strong{mv}
#'
#' @keywords splitting
#'
#' @export
#'
#' @examples
#'
#' if (interactive()) {
#'
#' # Make a 'csv' file
#'
#' data(mtcars)
#' tmpf = tempfile(fileext = '.csv')
#' write.table(mtcars,tmpf , sep = ',', row.names = FALSE, quote = FALSE)
#'
#' # 1. Read data file step by step
#'
#' # 1.1 Get information on data
#' r = splitH(readpath = tmpf)                             # stage 1
#' class(r)                                                # function
#'
#' # 1.2 Read data iteratively                             # stage 2
#' a = r(rows = 11, dropcols = c('am', 'vs'))              # iter1 no original row numbers
#' b = r(rows = 11, dropcols = c('am', 'vs'), orn = TRUE)  # iter2 w. original row numbers
#' c = r(rows = 11, dropcols = c('am', 'vs'))              # iter3 the last subset
#' d = r(rows = 11, dropcols = c('am', 'vs'))              # iter4 stop! Return to stage 1
#' print(list(a, b, c))
#'
#' # 2. Read data file completely
#'
#' r = splitH(readpath = tmpf)                             # stage 1
#' n = ceiling(32/13)                                      # read 13 rows at a time
#' a = replicate(n, r(rows = 13), simplify = FALSE)        # read file
#' class(a)                                                # list
#' print(a)                                                # a list of tables
#'
#' tmpf1 = tempfile(fileext = '.csv')                      # new location
#'
#' # 3. Iteratively write to new location
#'
#' r = splitH(readpath = tmpf, writepath = tmpf1)          # stage 1
#' n = ceiling(32/11)                                      # 11 rows each time
#' invisible(
#'  replicate(n, r(rows = 11) , simplify = FALSE)          # write to new location
#'  )
#' a = data.table::fread(tmpf1)                            # check result
#' dim(a)
#' print(head(a))
#'
#' unlink(tmpf1)
#'
#' tmpf2 = tempfile(fileext = '.csv')                      # new location
#'
#' # 4. Move file from tmpf to another location
#'
#' r = splitH(readpath = tmpf, writepath = tmpf2)          # stage 1
#' r(rows = Inf, how = mv, print = FALSE)                  # move to new location
#' a = data.table::fread(tmpf2)                            # check result
#' print(head(a))
#'
#' unlink(tmpf)
#' unlink(tmpf2)
#'
#' }
#'

splitH = function(readpath, writepath = NULL) {
                      rows = 'how' = tot = i = nmes = j = '<<-' = r = NULL
                  on.exit(i <- 0L, add = TRUE)
                  readpath = if (file.exists(readpath)) {normalizePath(readpath)
                             } else {stop('no file at source location!', call. = FALSE)}
                      info = match.fun(info, descend = FALSE)
                      dims = info(readpath)
                       tot = dims[[1L]]; nmes = dims[[2L]]
                         j = as.call(c(quote(list), list(srn = srn), eval(rename(nmes))))
                         r = data.table(NULL)
                function(rows, seq = TRUE, dropcols = NULL, how, print = TRUE, orn = FALSE) {
                      skip = eval(jump)
                        dt = if (skip > tot || is.infinite(rows) && i > 0L) {
                                stop('data reading was completed once!\n', call. = FALSE)
                            } else {
                                   if (isTRUE(print)) cat('\n\nsubset', i + 1L, ':\n')
                             fread(file = readpath, skip = skip, nrows = rows, col.names = nmes
                                 )[, j, env = list(j = j)] }
                     if (!is.null(dropcols)) dt[, let(dropcols, NULL), env = list(dropcols = substitute(dropcols))]
                     if (isTRUE(seq)) dt else dt = local({r <<- rbindlist(list(r, dt), use.names = TRUE)})
                        i <<- i + 1L
                     if (!is.null(writepath)) {
                         if (isTRUE(print)) print(dt)
                         if (is.infinite(rows)) {
                               how = substitute(how)
                              flag = if (how == 'scp') '-p' else '-b'
                         system2(how, c(flag, '-v', readpath, path.expand(writepath)), stdout = TRUE)
                         } else {
                           if (!seq) {
                              i <<- i - 1L
                    stop('when writing data, \"seq\" argument must be TRUE', call. = FALSE)}
                        if (isTRUE(orn)) dt[, let('srn', NULL)]
                        invisible(fwrite(dt, writepath, append = TRUE))
                    on.exit(message('\n\ndata have been written to ', writepath, '\n'), add = TRUE)
                 }} else return(dt)
            }}

#'@title Calculate The Number Of Combinations Of String Covering
#'
#'@description Calculates the cardinality of substring family (\emph{the covering}) and returns valid elements.
#'
#'@param x integer, length = 1 or character vector, length > 1, or character string
#'@param valid. logical, default, FALSE. When TRUE, valid substrings are returned. Requires character or string
#'@param wplot logical, FALSE. When TRUE, displays a plot of combinations versus number of characters in string
#'
#'@details This function finds the total, the minimum and the maximum number of combinations of individual string
#' covering or returns a subset of valid substrings (i.e. elements of the string) by retaining even transpositions
#' inside the combinations matrix. Very long strings may result in memory allocation error when \code{valid. = TRUE}.
#' Use [fcommon] instead.
#'
#' NOTE 1: Herein, the \emph{covering} represents the substring family that preserves the order of characters inside
#' the string. Hence, the use of combinations instead of partial permutations, \code{n!/(n-r)!}.
#'
#' NOTE 2: When \code{x} is an integer associated with an existing string, \code{x} must equal either \code{nchar(string)}
#' or \code{length(character)}.
#'
#'@returns When \code{valid. = FALSE} a prettified named vector showing the total, minimum and maximum number of
#' combinations for respective string covering. When \code{valid. = TRUE}, a character vector of valid substrings
#' longer than 2 characters each. In both cases, a plot is returned when \code{wplot = TRUE}.
#'
#'@seealso [fcommon]
#'
#'@keywords Proteomics
#'
#'@export
#'
#'@examples
#'
#' if (interactive()) {
#'
#' # 1. x is a character
#'
#' # 1.1 A string
#' x = 'tyrvvsvltvlhqdwlngkeykck'
#'
#' a = cover(x)
#' print(a)
#'
#' # 1.2 A vector
#' x = strsplit(x, '')[[1]]
#'
#' b = cover(x)
#' print(b)
#'
#' # 2. An integer
#'
#' n = length(x)
#' c = cover(n)
#' print(c)
#'
#' # 3. Valid substrings with plot
#'
#' d = cover(x, valid. = TRUE, wplot = TRUE)
#' print(head(d, 30))
#'
#' }
#'
#'

cover = function(x, valid. = FALSE, wplot = FALSE) {
         N = if (is.character(x)) {
             if (length(x) == 1L && isTRUE(nzchar(x)) && nchar(x) > 1L) {nchar(x)}
             else if (length(x) > 1L) {length(x)}
         } else if (is.numeric(x)) as.integer(x)
         xa = 2:(N-1)
          a = lapply(xa, \(i) comboGeneral(N, i))
         xx = sapply(a, nrow)
         on.exit(if (wplot) eval(plotish))
         if (!valid.) {
          b = structure(c(as.double(sum(xx))
                        , min(xx)
                        , max(xx))
                        , names = c('total', 'min', 'max'))
             format(b, scientific = TRUE, justify = 'none', digits = 3)
         } else {
           if (is.numeric(x)) stop('numeric quantities are not validated!', call. = FALSE)
            xv = if (length(x) == 1L && nchar(x) > 1L) strsplit(x, '')[[1L]] else x
             v = x; mn = NULL
          for (n in seq(along=a)) {
                      mn = a[[n]][c(1L, eval(cond)),]
                       v = c(v, apply(eval(cmm), 1L, paste0, collapse = ''))
          }
      unique(v[nchar(v) > 1L]) |> sort()
    }
}

#' @title Locate And Extract Modifications Or Monoisotopic Masses From A Modified Peptide
#'
#' @description Finds and tabulates amino acid modification sites and extracts modifications or monoisotopic masses
#' from modified peptide data representation.
#'
#' @param string character, length 1. Modified or unmodified peptide, or NULL
#' @param wrap character, length 1. The closing (right-hand) side of any of the bracket types ']', ')', '\}' that wrap
#'   the modifications, such as in protein mass spectrometry data representation of modified peptides. Default, ']'
#' @param inbracket character, length 1. Same as above for brackets used inside modification wrappings. Default, ')'
#' @param except character, length >= 1. Default, NULL. Punctuation marks or characters that appear along modifications
#'   and are needed to remain present in the output: '-', '+', ',', ';', ':', '=', '.', `[:digit:]`, `[:alpha:]`, '\\w+', ' '
#' @param rmve character, length 1. Default, NULL. Regular expression. Digits or extra characters that need to be removed from
#'   the output (see Examples)
#'
#' @details Although capable of handling most situations, it is recommended that the \code{wrap}ping bracket type
#' remains consistent throughout and the \code{inbracket} type \emph{be different} from \code{wrap}ping type.
#' No extra characters are removed from result when \code{except = rmve = NULL}.
#'
#' This utility covers most data representation styles for modified peptide. However, clean data results are not guaranteed.
#' The template for letter casing accepted for modified peptide and for modifications should match those presented in Examples:
#' upper case for peptide and mixed case for modifications.
#'
#' @return A 'data.table' class data frame containing the unmodified peptide, the modified peptide, the modification site
#' (i.e. the amino acid code letter and location inside the peptide) and the associated modification(s). In case of monoisotopic
#' mass extraction, monoisotopic mass values populate column "Modification" as "character" types. Multiple modifications (identical
#' or not) found at the same site are listed as many times as they appear at that site. Unmodified, endogenous peptides are listed
#' with no other information. Empty strings are listed as such with a warning.
#'
#' @seealso \link[base]{regex}
#'
#' @keywords Proteomics
#'
#' @export
#'
#' @examples
#'
#' if (interactive()) {
#'
#' # Completely made-up modified peptides:
#'
#' # 1. Modifications
#'
#' # 1.1 Default brackets
#' string = 'K[Prop_A][Met][Prop (C)]PSSABCELR[Prop][Prop][Prop]FQC[Carba (C)]GQQ[Met +44]TARP'
#'
#' a = locateMod(string)
#' print(a)                                                              # with extra-characters
#' b = locateMod(string, except = '\\w+', rmve = '(\\(.*\\)|_[A-Z]|[0-9])')
#' print(b)                                                              # without extra-characters
#'
#' # In this example argument "rmve" contains the default in-brackets
#'
#' # 1.2 Alternative bracketing
#'
#' string = 'K{Prop_A}{Met}{Prop [A]}PSSABCELR{Prop +15}{Prop}{Prop}FQC{Carba [C]}GQQ{Met +44}TARP'
#'
#' c = locateMod(string, '}', ']')
#' print(c)
#' d = locateMod(string, '}', ']', except = '\\w+', rmve = '(\\[.*\\]|_[A-Z]|[0-9])')
#' print(d)
#'
#' # In this example argument "rmve" contains the alternative in-brackets
#'
#' # 2. Empty string
#'
#' empty = locateMod(""); print(empty)
#'
#' # 3. Monoisotopic masses
#'
#' string = 'TAAC[+57.021464]PPC[+57.021464]PAPPAPS[+162.052824]VFLTLMISR'
#' e = locateMod(string)
#' print(e)                                                             # with extra-characters
#' f = locateMod(string, rmve = '[[:punct:]]')$Modification
#' print(f)                                                             # incorrect values
#' g = locateMod(string, rmve = '\\+')$Modification
#' print(g)                                                             # correct!
#' class(g)                                                             # character
#' }
#'

locateMod = function(string, wrap = "]", inbracket = ')', except = NULL, rmve = NULL) {
                     string = if (isTRUE(nzchar(string) && !identical(' ', string))) {as.character(string)
                              } else {warning('empty string!', call. = FALSE, noBreaks. = TRUE)}
                       wrap = match.arg(wrap, choices = c("]", ")", "}"))
                   brackets = as.character(c('}','{',']','[',')','('))
                      awrap = brackets[which(brackets %in% wrap) + 1L]
                   abracket = brackets[which(brackets %in% inbracket) + 1L]
                    if (!exists('plac', mode = 'character', inherits = FALSE)) plac = eval(.plac)
                         br = regmatches(string, gregexpr(plac, string))[[1L]]
               if (length(br) == 0L) {
                         dt = data.table(
                                             'Peptide' = string
                                           , 'Modified Sequence' = string
                                           , 'AA' = NA_character_
                                           , 'Location' = NA_integer_
                                           , 'Modification' = NA_character_
                                        )
                  } else {
                         uu = eval(.uu); u = eval(.u); uv = eval(.uv)
                      modif = regmatches(string, gregexpr(u, string, perl = TRUE))[[1L]]
                      modif = gsub(uv, '', modif)
                      delayedAssign('pep', paste0(
                                          regmatches(string, gregexpr(u, string, perl = TRUE), invert = TRUE)[[1L]]
                                        , collapse = ''))
                      if (!is.null(except)) {
                      if (!exists('ARGS', mode = 'character', inherits = FALSE)) ARGS = eval(.ARGS)
                                 except = match.arg(except, arg = ARGS, several.ok = TRUE)
                                      w = eval(.w)
                                  modif = gsub(w, '', modif)}
                      if (!is.null(rmve)) modif = gsub(rmve, '', modif)
                      delayedAssign('n.mod', length(modif))
                         sp = strsplit(string, split = '', FALSE, FALSE, FALSE)[[1L]]
                         wp = grep(wrap, brackets)
                     startx = paste0('(', '\\', brackets[wp + 1L], ')', '+', collapse = '')
                       endx = paste0('(', '\\', brackets[wp], ')', '+', collapse = '')
                      start = grep(startx, sp) - 1L
                        end = grep(endx, sp); l.en = length(end)
             if (any(diff(end) == 1L)) {
                               diffs = c(FALSE, diff(end) == 1L)
                              wnodif = which(!diffs)
                               start = start[wnodif]; end = end[wnodif]}
                        end = c(0L, end[-l.en])
                      AAloc = cumsum(start - end)
                    AAmodif = sp[start]
                       site = which(AAmodif %in% wrap); l.s = length(site)
          while (l.s) {
             AAmodif[site] <- AAmodif[site - 1L]
                       l.s <- l.s - 1L}
                         dt = try(data.table(
                                             'Peptide' = rep(pep, n.mod)
                                           , 'Modified Sequence' = rep(string, n.mod)
                                           , 'AA' = AAmodif
                                           , 'Location' = AAloc
                                           , 'Modification' = modif
                                        ))
            }
             return(dt)
       }

#' @title Find Substring Locations Inside A String
#'
#' @description Finds all locations of a known character substring inside a character string.
#'
#' @param subchain character, length 1, e.g. a peptide sequence
#' @param chain (named) character, length 1 or a (named) list of such characters such as a
#'   list of protein chains obtained from a \emph{fasta} file
#' @param outlist logical. Default, FALSE, the output is a (named) integer vector of locations. Otherwise,
#'   it is a (named) list of location vectors, each corresponding to a chain in a list of chains
#' @param named logical. Default, FALSE. Output is not named. Otherwise, the output is named
#' @param all. logical, default FALSE, returns the leftmost or the rightmost location inside the chain.
#'   When TRUE, returns all locations inside the chain for each chain in a list
#' @param which symbol. Location to report. Default, \strong{min}. Requires \code{all. = FALSE}. \code{which = min}
#'   returns the leftmost location inside the chain, e.g. closest to the N-terminus of a protein chain. \code{which = max}
#'   returns the rightmost location, e.g. closest to the C-terminus of a protein chain. Overwritten when \code{all. = TRUE}
#' @param ignore.case,perl,fixed,useBytes arguments to [base::gregexpr]
#'
#' @details Wrapper to [base::gregexpr], the function scans all chains in a list of chains to find subchain locations.
#' The location is defined as the position inside the chain relative to the left end of the chain of the first subchain character.
#'
#' @returns A (named) integer or a (named) list of integer vectors of \code{subchain} locations inside the \code{chain}.
#'
#' @seealso \link[base]{gregexpr}
#'
#' @keywords Proteomics
#'
#' @export
#'
#' @examples
#'
#' if (interactive()) {
#'
#' # 1. List of chains
#'
#' chain = list(chain1 = 'alpdxoipoyloiekladxoipoylyl',
#'              chain2  = 'kdxoipoylyydxoipoylopldxoipoylac')
#' subchain = 'DXOIPOYL'                                                  # ignoring the case
#'
#' findLoc(subchain, chain, outlist = TRUE, named = TRUE, all. = TRUE)    # named list
#' findLoc(subchain, chain, named = TRUE, all. = TRUE)                    # named integer
#' findLoc(subchain, chain, outlist = TRUE, named = TRUE)                 # the leftmost positions
#' findLoc(subchain, chain, which = max, named = TRUE)                    # the rightmost positions
#'
#' # 2. Single chain
#'
#' chain = chain[[1]]
#'
#' findLoc(subchain, chain, all. = TRUE)
#' findLoc(subchain, chain, which = max)
#' findLoc(subchain, chain)                                               # default location
#' findLoc(subchain, chain, which = max, ignore.case = FALSE)             # not ignoring the case
#'
#' }
#'

 findLoc = function(subchain, chain, outlist = FALSE, named = FALSE, all. = FALSE
                  , which = min, ignore.case = TRUE, perl = FALSE, fixed = FALSE, useBytes = FALSE) {
                       nc = if (isFALSE(named)) NULL else names(chain)
                    chain = if (isTRUE(all(nzchar(chain)))) as.character(chain)
                 subchain = if (isTRUE(nzchar(subchain))) as.character(subchain)
                 if (isTRUE(nchar(chain) < nchar(subchain))) {stop('the chain is too short!', call. = FALSE)
                } else {
                 which = substitute(which)
                 which = match.fun(match.arg(as.character(which), choices = c('min', 'max')), descend = FALSE)
                     w = listenv(); v = listenv()
                     v = gregexpr(subchain, chain, ignore.case, perl, fixed, useBytes)
              names(v) = nc
                     w = lapply(v, \(i) `[`(i, i > 0L)); v <- NULL
                     w[lengths(w) == 0L] <- NULL
                 if(isFALSE(all.)) {
                   if(isFALSE(outlist)) vapply(w, \(i) which(i), integer(1L)) else lapply(w, \(i) which(i))
                 } else {
                    if(isFALSE(outlist)) unlist(w, TRUE, TRUE) else w}}
        }

test_that("returns named list", {

                chain = list(chain1 = 'alpdxoipoyloiekladxoipoylyl',
                            chain2  = 'kdxoipoylyydxoipoylopldxoipoylac')
                           subchain = 'DXOIPOYL'
                   ll = findLoc(subchain, chain, outlist = TRUE, named = TRUE, all. = TRUE)
                   expect_identical(is.list(ll), TRUE)
                   expect_named(ll, c('chain1', 'chain2'))

})

#'@title Find Substring Combinations
#'
#'@description
#'Internal function for \"fcommon\" - the fast version of function \"common\".
#'
#'@param u character, length 1. A string
#'@param rows integer length 1. Default 100. The number of rows (combinations) in each
#'  combinations matrix
#'@param ... not used
#'
#'@keywords internal
#'@noRd
#'

core = function(u, rows = 100, ...) {
        b = strsplit(as.character(u), '')[[1L]]
       nc = nchar(u)
        w = 2L:(nc - 1L)
  listenv = match.fun(listenv, descend = FALSE)
      cbp = listenv()
        m = character()
  for (i in seq(along=w)) {
      A = comboIter(v = b
                    , m = w[i]
                    , FUN = function(x, chain = u, ...) {
                            y = paste0(x, collapse = '', ...)
                            if (eval(cond)) y else NULL
                    }
                    , FUN.VALUE = list(1L)
                    , ...
      )
        N = ceiling(as.double(comboCount(b, w[i]))/rows)
      for (j in 1:N) {
        P = A$nextNIter(rows)
        m <- unique(c(m, P[lengths(P) > 0L])); P <- NULL
        j = j + 1
      }
      A@startOver()
       cbp[[i]] = m; m <- NULL
  }
  unlist(cbp)
}

#' @title One-hot Encoder And Decoder Of Variables
#'
#' @description Encodes logical, categorical, integer and double type variables.
#'
#' @param x a (named) vector or list for encoding. Missing data are removed. For decoding, a dense or sparse
#'   matrix (preferably, the result of encoding) representing a single source data column
#' @param type symbol. Choices: \strong{encode} - one-hot encoding, \strong{decode} - revert to original
#' @param omc character length 1. \strong{O}utput \strong{m}atrix \strong{c}lass. Default, 'dgCMatrix', other
#'   option, 'matrix'
#' @param verbose logical, default TRUE, display messages
#'
#' @details This utility one-hot encodes when \code{type = encode} and verifies the encoded result (or any
#' matrix of encodings obtained with [getEV] extractor) when \code{type = decode}. It detects illicit states.
#'
#' @returns Encoding returns a matrix of \code{length(x)} rows and \code{length(unique(x))} columns or a warning.
#' Decoding returns a (named) vector or a warning. List vectors are returned unlisted. Integer(ish) vectors,
#' converted to integer, character vectors - to factor, double or logical vector types remain unchanged.
#'
#'@keywords conversion
#'
#'@export
#'
#'@examples
#'
#' if (interactive()) {
#'
#' # 1. Encode type "double"
#'
#' x = runif(9)                           # numeric, length 9
#' names(x) = letters[1:9]                # named
#' typeof(x)
#' a = oneHot(x, encode)                  # a sparse matrix of "dgCMatrix" class
#' b = oneHot(a, decode)                  # a type "double" named numeric, length 9
#' isTRUE(all.equal(x, b))                # TRUE
#' typeof(b)
#' print(x); print(b)
#'
#' # 2. Type "logical" with missing values
#'
#' y = c(TRUE, TRUE, NA, FALSE, TRUE, NA) # logical, length 6 with missing values
#' typeof(y)
#' a = oneHot(y, encode, 'matrix')
#' print(a)                               # a dense matrix
#' b = oneHot(a, decode)                  # revert
#' all.equal(y, b)                        # missing values in y removed
#' typeof(b)
#' print(x); print(b)
#'
#' # 3. iris data
#'
#' data(iris)
#' a = lapply(iris, oneHot, encode)       # encode entire data
#' b = as.data.frame(
#'          lapply(a, oneHot, decode)     # revert
#'      )
#' identical(iris, b)                     # TRUE. Now, replace iris data with
#'                                        # mtcars data!
#'
#' # 4. Illicit states in one-hot encoding
#'
#' `3.41` = c(1,0,0,1,1,0,0,1)            # encoded type "double"
#' `0.12` = c(0,1,0,0,0,1,1,0)
#'  a = cbind(`3.41`, `0.12`)             # form encoded matrix
#'  print(a)                              # matrix resembling one-hot encoding
#'  x = oneHot(a, decode)                 # illicit state detected
#'  print(x)                              # list with 2 different data types
#'
#'}
#'

oneHot = function(x, type, omc = 'dgCMatrix', verbose = TRUE) {
                     type = substitute(type)
                     type = match.arg(as.character(type), choices = c('encode', 'decode'))
                      omc = match.arg(omc, choices = c('matrix', 'dgCMatrix'))
               switch(type
                    , encode = {
                     if (!is.null(dim(x))) {
                      stop('data should have one dimension only!', call. = FALSE)}
                     else {
                       delayedAssign('sumsetc', sum(m) == n.x && all(rowSums(m) == 1L))
                      if (anyNA(x) && verbose) cat('\nall missing values were removed!\n')
                        x = if (is.list(x)) unlist(x)[!is.na(x)] else x[!is.na(x)]
                      if(length(x) == 0L && verbose) {cat('\nno remaining values in input!!\n\n')
                    } else if (is.double(x) && verbose) {cat('encoding type \"double\"!...\n')
                    } else if(verbose) {cat('encoding type \"integer(ish)\"...\r')}
                     rn.x = names(x)
                        x = fdroplevels(if (is.factor(x)) x else as.factor(x))
                      n.x = length(x); ll = levels(x); n.l = length(ll)
                        m = eval(M)
                     for (i in 1:n.l) m[, i] = fmatch(x, ll[i], 0L, NULL)
                     if (isTRUE(sumsetc)) return(as(m, omc)) else warning('ambiguous encoding!\n', call. = FALSE)
                }}
                   , decode = {
                    if (is.null(dim(x))) {stop('data must have at least 2 columns', call. = FALSE)
                    } else {
                     delayedAssign('sms', rowSums(x) != 1L)
                     delayedAssign('any0len', any(sapply(out, length) == 0L))
                     delayedAssign('nolenout', length(out) != nrow(x))
                    if (any(sms) && verbose) cat('\nillicit state at row(s) ', which(sms), ' in input!\n')
                    rn.x = dimnames(x)[[1L]]
                     n.x = dimnames(x)[[2L]]
                       g = match.fun(g, descend = FALSE)
                  if (eval(tof) && verbose) {cat('decoding type \"double\"...\n')
                  } else if (verbose) (cat('decoding type \"integer(ish)\"...\r'))
                     out = apply(x, 1L, g, n.x, simplify = TRUE)
                     out = structure(type.convert(out, as.is = FALSE, numerals = 'no.loss'), names = rn.x)
                     if (nolenout || any0len && verbose) cat('ambiguous decoding!\n')
                  return(out)
              }}, character(0))
          }

#' @title Identify Common Substrings In A Pair Of Strings
#'
#' @description Checks and identifies substrings that are common in a pair of strings.
#'
#' @param X,Y character, length = 1 each: a string, such as a protein chain. When \code{Y = NULL},
#'    all arguments coming after \code{X} should be named
#' @param from,to integer length 1 each. Range of string character lengths to be identified. When \code{to}
#'    is missing from call, only substrings of length equal to \code{from} value are sought
#' @param lower,upper integer, length 1 each, default NULL: spans all substring combinations. Otherwise, spans
#'    only a subset of combinations starting with the \code{lower}-th and ending with the \code{upper}-th
#'    combinations. See [RcppAlgos::comboGeneral]
#' @param outlist logical. Default, FALSE, return a character vector of common substrings. TRUE, return a list
#'    of valid substrings found in each chain
#' @param rows integer length 1. Default 1000. The number of rows (combinations) in each combinations matrix
#' @param wait integer length 1. Default 100. Duration of the background process polling. Unit: milliseconds.
#'    During this time, the main process is blocked.
#' @param ... not used
#'
#' @details Using brute combinatorial approach, this utility partitions each \code{chain} in the pair into all possible
#' combinations of substrings with lengths inside the \code{from}-\code{to} window. Each combination is filtered for
#' uniqueness and appartenance to its original chain (i.e. for validity), yielding a fraction of common substrings out of all
#' substrings. Inside the valid set, each substring represents the \emph{truncation} in a sequence of longer substrings.
#'
#' The overhead causes severe search time increase with the total number of combinations which, depend on \code{chain} and
#' substring lengths, and search window width. Slight performance improvement comes from variations in \code{rows} value
#' i.e. by balancing the number of combinations matrix \code{rows} with the total number of combinations matrices. The default
#' value is an acceptable starting point. Chains of several hundreds of characters, such as long protein chains, may take hours
#' on some machines. In such cases, \code{rows} value can be risen toward the order of 1e6. When \code{Y != NULL}, the partitioning
#' of chains runs asynchronously.
#'
#' The \code{lower} and \code{upper} limits. Setting any or both of these arguments reduces the search time and only establishes
#' whether \emph{any} common substrings exist at all, without guaranteeing complete results. In case of very long chains, setting
#' these limits too wide apart may lead to memory allocation error. Checking the number of combinations before run is
#' recommended, see [RcppAlgos::comboCount], [utils::combn] etc.
#'
#' NOTE: This utility uses background processing. Check "Security Considerations" in \strong{callr} package documentation.
#'
#' @returns A character vector of common substrings or a list of valid substrings found in both chains. When \code{Y = NULL}, a list
#' of all valid substrings in \code{X} within the \code{from}-\code{to} window.
#'
#' @seealso \link[RcppAlgos]{comboIter}, \link[RcppAlgos]{comboGeneral}, \link[RcppAlgos]{comboCount}
#'
#' @keywords Proteomics
#'
#' @export
#'
#' @examples
#'
#' if (interactive()) {
#'
#' # 1. A set of chains
#'
#'  X = 'alpdxoipoyloiekladxoipoylyl'
#'  Y = 'kdxoipoylyydxoipoylopldxoipoylac'
#'
#' # 1.1 Identify all 4-character common substrings
#'  system.time(a <- common(X, Y, 4))
#'  print(a)                                              # common substrings set
#'
#'  # 1.2 Check the commonality of substrings in "a"
#'  chain = list(chain1 = 'alpdxoipoyloiekladxoipoylyl',
#'               chain2 = "kdxoipoylyydxoipoylopldxoipoylac")
#'
#'  b = sapply(a, findLoc, chain, TRUE, TRUE, TRUE, simplify = FALSE)
#'  print(b)                                              # a named list
#'
#'
#'  any(lengths(b) == 0L)                                 # FALSE
#'  identical(length(a), length(b))                       # TRUE
#'
#' # 2. Combinatorial Subset
#'
#'  c = common(X, Y, from = 4, to = 9, lower = 5, upper = 100, outlist = TRUE)
#'  print(c)                                              # named list
#'
#' # 3. Single Chain
#'
#'  d = common(X, from = 3, to = 5)
#'  print(d)                                              # valid substrings set
#'
#' }
#'

common = function(X, Y = NULL, from, to, lower = NULL, upper = NULL, outlist = FALSE, rows = 1000, wait = 100, ...) {
  X = as.character(X); if (!is.null(Y)) Y = as.character(Y)
  on.exit(ca <- NULL, add = TRUE); on.exit(cb <- NULL, add = TRUE)
  b = if (isTRUE(nzchar(X)) && isTRUE(nchar(X) > 1L) && isTRUE(length(X) == 1L)) {strsplit(X, '')[[1L]]
      } else {stop('X and Y must be strings!', call. = FALSE)}
  if (missing(to)) to <- from
  w = from:to
  if (!is.null(Y)) {
  prc = r_bg(function(w, listenv, comboIter, comboGeneral, rows, comboCount, Y, cond, lower, upper, ...) {
         bp = if (!is.null(Y) && isTRUE(nzchar(Y)) && isTRUE(nchar(Y) > 1L) && isTRUE(length(Y) == 1L)) {strsplit(Y, '')[[1L]]
              } else {stop('X and Y must be strings!', call. = FALSE)}
                   listenv = match.fun(listenv, descend = FALSE)
                       cbp = listenv()
                         m = character()
                  for (i in seq(along=w)) {
                    if (is.null(lower) && is.null(upper)) {
                                     A = comboIter(v = bp
                                                 , m = w[i]
                                                 , FUN = function(x, chain = Y, ...) {
                                                               y = paste0(x, collapse = '')
                                                               if (eval(cond)) y
                                                 }
                                                 , FUN.VALUE = list(1L)
                                                 , ...
                                                 )
                                    N = ceiling(as.double(comboCount(bp, w[i]))/rows)
                                    for (j in 1:N) {
                                               P = A$nextNIter(rows)
                                               m <- unique(c(P, m))
                                               j = j + 1
                                              }
                                   A@startOver(); P <- NULL
                    } else {
                      A = comboGeneral(v = bp
                                       , m = w[i]
                                       , FUN = function(x, chain = Y, ...) {
                                         y = paste0(x, collapse = '')
                                         if (eval(cond)) y
                                       }
                                       , FUN.VALUE = list(1L)
                                       , chain = Y
                                       , lower = lower
                                       , upper = upper
                                       , ...
                                       )
                      m <- unique(A); A <- NULL
                    }
          cbp[[i]] = m; m <- NULL
            }
     unlist(cbp)
  }, args = list(w, listenv, comboIter, comboGeneral, rows, comboCount, Y, cond, lower, upper, ...), user_profile = FALSE, supervise = TRUE)}
  cb = listenv()
   m = character()
           for (i in seq(along=w)) {
             if (is.null(lower) && is.null(upper)) {
                              A = comboIter(v = b
                                          , m = w[i]
                                          , FUN = function(x, chain = X, ...) {
                                                        y = paste0(x, collapse = '')
                                                        if (eval(cond)) y
                                                       }
                                          , FUN.VALUE = list(1L)
                                          , ...
                                          )
                             N = ceiling(as.double(comboCount(b, w[i]))/rows)
                             for (j in 1:N) {
                                       P = A$nextNIter(rows)
                                       m <- unique(c(P, m))
                                       j = j + 1
                                      }
                            A@startOver(); P <- NULL
             } else {
               A = comboGeneral(v = b
                                , m = w[i]
                                , FUN = function(x, chain = X, ...) {
                                  y = paste0(x, collapse = '')
                                  if (eval(cond)) y
                                }
                                , FUN.VALUE = list(1L)
                                , chain = X
                                , lower = lower
                                , upper = upper
                                , ...
               )
               m <- unique(A); A <- NULL
             }
                  cb[[i]] = m; m <- NULL
           }
  cb = unlist(cb)
  if(!is.null(Y)) {ca = if('ready' %in% prc$poll_io(wait)['process']) {prc$get_result()} else {
                                                                                               prc$wait()
                                                                                               prc$get_result()
                                                                                               }}
  if (!outlist && !is.null(Y)) intersect(ca, cb) else if (!is.null(Y)) list('X' = cb, 'Y' = ca) else cb
}

test_that("common strings are found", {

                          chain1 = 'alpdxoipoyloiekladxoipoylyl'
                         chain2  = 'kdxoipoylyydxoipoylopldxoipoylac'

                               a = common(chain1, chain2, 3)

                              expect_length(a, 8)
  })

#'@title Plot For Function "cover"
#'@keywords internal
#'@noRd
#'

plotish = substitute(expr = {
                   olp = par(no.readonly = TRUE)
                   par(list(mar = c(4,4,3,1), tcl = -0.1, ylog = TRUE))
                   plot(xa, xx, xaxt = 'n', yaxt = 'n', type = 'b', pch = 20, xlab = 'Characters', ylab = ''
                        , main = 'Covering Combinations', cex.main = 0.8)
                   axis(1, cex.axis = 0.7, xaxs = 'i', las = 1, at = seq(min(xa), max(xa), by = 4)
                        , labels = format(seq(min(xa), max(xa), by = 4)), cex.lab = 0.8)
                   axis(2, at = seq(min(xx), max(xx), by = 5e+5), , cex.axis = 0.7, yaxs = 'i', las = 1
                        , ylim = c(min(xx), max(xx)*1.05), labels = format(seq(min(xx), max(xx), by = 5e+5)
                        , scientific = TRUE, digits = 2))
                   par(olp)}
                  )

#' @title Fast Identify Common Substrings In A Pair Of Strings
#'
#' @description Checks and identifies substrings that are common to a pair of strings.
#'
#' @param x,y character, length 1 each: a string, such as a protein chain. \code{y} can be missing
#'
#' @details This utility identifies common substrings in the \code{x}, \code{y} pair of strings by isolating \emph{sequences} of
#' identical characters in both strings which then, are packed into substrings and validated. All one-character substrings are removed.
#' When \code{y} is missing, \code{x} is cleaved at each letter producing all substrings longer than 2 characters.
#'
#' @returns A sorted character vector of common substrings of length >= 2 characters each. When \code{y} is missing from call, a sorted
#' character vector of valid substrings in \code{x} of length >= 2 characters each.
#'
#' @keywords Proteomics
#'
#' @seealso [cover]
#'
#' @export
#'
#' @examples
#'
#' if (interactive()) {
#'
#'  # 1. Check for common substrings in the pair below
#'
#'  x = 'dvvmtqsplslpvtpgepasiscrssqslaktyrvvsvltvlhqdwlngkeykckvv'
#'  y = 'mtqspltyrvvsvltvlhqdwlngkeykcksnkalpapiektisk'
#'
#' # 1.1 Common substrings
#'  system.time(a <- fcommon(x, y))
#'  print(head(a, 30))
#'
#' # 1.2 Cleaving (slow on very long strings!)
#'  system.time(aa <- fcommon(x))
#'  system.time(bb <- fcommon(y))
#'
#' # 1.3 Identical results
#'  A = sort(intersect(aa, bb))                                # common substrings
#'  identical(a, A)                                            # TRUE
#'
#' # 2. Different methods for valid substrings
#'
#' x = 'tyrvvsvltvlhqdwlngkeykck'
#'
#' # 2.1. Combinations matrix (slower!)
#' system.time(am <- cover(x, valid. = TRUE))                  # valid substrings
#'
#' # 2.2 String cleaving
#' system.time(ac <- fcommon(x))                               # valid substrings
#'
#' identical(am, ac)                                           # TRUE
#'
#'}
#'

fcommon = function(x, y) {
                 frec = match.fun(frec, descend = FALSE)
                 seqv = match.fun(seqv, descend = FALSE)
              on.exit(rez <- NULL, add = TRUE)
              if (missing(y)) {
               if (!nzchar(x)) stop('\nstring should not be empty!\n', call. = FALSE)
                            rez = do.call(frec, list(x))
                            rez = sapply(rez, paste0, collapse = '', USE.NAMES = FALSE); outl <- NULL
                            rez = rez[which(sapply(rez, grepl, x, useBytes = TRUE, simplify = TRUE))]
                } else {
                    x = as.character(x); y = as.character(y)
              if (!nzchar(x) || !nzchar(y)) stop('\nno string should be empty!\n', call. = FALSE)
                   fo = list(x, y)
                  m.l = which.min(c(nchar(fo[[1L]]), nchar(fo[[2L]])))
                   xv = strsplit(fo[[m.l]], split = '')[[1L]]
                   yv = strsplit(fo[[-m.l]], split = '')[[1L]]; fo <- NULL
                   xv = append(xv, rep(NA_character_, abs(length(yv) - length(xv))))
                    z = zz = yv
                    m = cbind(xv, z, zz)
                   k0 = m[seqv(eval(ii)), 1L]; kk0 = m[seqv(eval(jj)), 1L]
                   kl = function() {
                                  k = list(); kk = list()
                                  N = max(length(xv), length(yv))
                           for (i in 1:N) {
                                        z <<- shift(z, -1L)
                                       zz <<- shift(zz, 1L)
                                        m <<- cbind(xv, z, zz)
                                     k[[i]] = m[seqv(eval(ii)), 1L]
                                    kk[[i]] = m[seqv(eval(jj)), 1L]
                                    }
                                kl = c(k0, kk0, k, kk) |> unique()
                                xv = yv = m = k0 = k = kk0 = kk <- NULL
                                kl[lengths(kl) > 1L]
                                }
                            outl = kl()
                              zu = sapply(outl, paste0, collapse = '', USE.NAMES = FALSE); outl <- NULL
                              zv = sapply(zu, frec)
                              zv = unlist(zv) |> unique()
                             rez = intersect(
                                      zv[which(sapply(zv, grepl, x, useBytes = TRUE, simplify = TRUE))]
                                    , zv[which(sapply(zv, grepl, y, useBytes = TRUE, simplify = TRUE))]
                            )
                      }
            sort(rez)
      }

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getEncodedVar.R
\name{getEV}
\alias{getEV}
\title{Extract Encoded Variables From Encoded Split Or Tiled Data}
\usage{
getEV(en, name, ...)
}
\arguments{
\item{en}{a (named) list, or \link[listenv]{listenv} of matrices or a single matrix, all containing multiple
encoded variables. See \link{oneHot} decoder for lists and matrices containing single encoded variables}

\item{name}{character, length 1. Column name as found in source data}

\item{...}{default, empty. Used to convert the class of extracted matrix to 'dgCMatrix' or 'matrix'}
}
\value{
A dense or sparse matrix of single encoded variable which can be decoded with the \link{oneHot} decoder.
}
\description{
Extracts a single encoded variable from a \link[base]{list} or \link[listenv]{listenv}
of encoded matrices containing multiple encoded variables
}
\details{
This function includes code from package "Matrix.utils" v 0.9.8, published under GPL-3
license, currently removed from CRAN. With thanks to the package Author!

NOTE 1: If \code{name} is a source data column name that appears inside other column names, the extracted matrix
will combine all encoded matrices having this name inside their column names. Although the extracted matrix is a
proper matrix of encodings, it no longer represents \emph{a single} encoded data column. As result, upon decoding,
the \link{oneHot} decoder will report ambiguous decoding.

NOTE 2: a warning reading either \emph{"single-column encoded matrix for ..."} or \emph{"number of columns of
result is not a multiple of vector length (arg 2) ..."} may appear when extracting an encoded categorical variable
from a list of encoded matrices. Most likely, this happens with low cardinality encoded variables. The warning
signals that most encoded matrices associated with respective variable contain subsets of only one category (level)
when, ideally, most of these matrices should contain a mixture of two or more categories or levels; thus, allowing
matrix row-binding by category's label. One or more of the following suggestions will solve the issue: a) shuffle
the data before encoding, b) increase the number of \code{rows} in data chunks when encoding, c) if memory allows,
opt for \link{tileHot} encoding single matrix output, as shown in Example 2.1, solution c.
}
\examples{

if (interactive()) {

# 1. mtcars data have all columns type "double"

data(mtcars)
a = lapply(mtcars, oneHot, encode)                       # encode mtcars data
print(a)                                                 # list of sparse matrices
b = getEV(a, 'cyl')                                      # extract encoded "cyl" column
print(b)                                                 # a 32x3 sparse matrix
c = oneHot(b, decode)                                    # revert
identical(mtcars$cyl, c)                                 # FALSE. 'mtcars$cyl' is type "double"
isTRUE(all.equal(mtcars$cyl, c))                         # TRUE

# 2. Warnings associated with low cardinality categorical variable

# See tileHot() Examples for full decoding of a dataset

# 2.1 Make 'csv' file
data(iris)                                               # low cardinality "Species"
tempf = tempfile(fileext = '.csv')
write.table(iris, tempf , sep = ',', row.names = FALSE, quote = FALSE)

A = tileHot(readpath = tempf, rows = 14, splits = 3)     # encoded tiles list
print(A[[11]][[5]])                                      # e.g. one-column matrix
a = getEV(A, 'Species')                                  # warning
colSums(a)                                               # incorrect!

# solution b
B = tileHot(readpath = tempf, rows = 60, splits = 3)     # increase number of rows
b = getEV(B, 'Species')                                  # still warning
colSums(b)                                               # incorrect!

# Solution b) could work in combination with solution a)

# solution c
C = tileHot(tempf, rows = 14, splits = 3, orn = TRUE)    # encoded matrix
c = getEV(C, 'Species')                                  # no warning
colSums(c)                                               # correct!

unlink(tempf)

# 2.2 Shuffled 'csv' file
tempf = tempfile(fileext = '.csv')
iris22 = iris[{ set.seed(327); sample.int(150) },]      # shuffled iris data
write.table(iris22, tempf , sep = ',', row.names = FALSE, quote = FALSE)

A = tileHot(readpath = tempf, rows = 14, splits = 3)    # same as above

#solution a
a = getEV(A, 'Species')                                 # no warning
colSums(a)                                              # correct!

unlink(tempf)

}


}
\seealso{
\link{oneHot}, \link{tileHot}
}
\keyword{extractor}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/common.R
\name{common}
\alias{common}
\title{Identify Common Substrings In A Pair Of Strings}
\usage{
common(
  X,
  Y = NULL,
  from,
  to,
  lower = NULL,
  upper = NULL,
  outlist = FALSE,
  rows = 1000,
  wait = 100,
  ...
)
}
\arguments{
\item{X, Y}{character, length = 1 each: a string, such as a protein chain. When \code{Y = NULL},
all arguments coming after \code{X} should be named}

\item{from, to}{integer length 1 each. Range of string character lengths to be identified. When \code{to}
is missing from call, only substrings of length equal to \code{from} value are sought}

\item{lower, upper}{integer, length 1 each, default NULL: spans all substring combinations. Otherwise, spans
only a subset of combinations starting with the \code{lower}-th and ending with the \code{upper}-th
combinations. See \link[RcppAlgos:combinatoricsGeneral]{RcppAlgos::comboGeneral}}

\item{outlist}{logical. Default, FALSE, return a character vector of common substrings. TRUE, return a list
of valid substrings found in each chain}

\item{rows}{integer length 1. Default 1000. The number of rows (combinations) in each combinations matrix}

\item{wait}{integer length 1. Default 100. Duration of the background process polling. Unit: milliseconds.
During this time, the main process is blocked.}

\item{...}{not used}
}
\value{
A character vector of common substrings or a list of valid substrings found in both chains. When \code{Y = NULL}, a list
of all valid substrings in \code{X} within the \code{from}-\code{to} window.
}
\description{
Checks and identifies substrings that are common in a pair of strings.
}
\details{
Using brute combinatorial approach, this utility partitions each \code{chain} in the pair into all possible
combinations of substrings with lengths inside the \code{from}-\code{to} window. Each combination is filtered for
uniqueness and appartenance to its original chain (i.e. for validity), yielding a fraction of common substrings out of all
substrings. Inside the valid set, each substring represents the \emph{truncation} in a sequence of longer substrings.

The overhead causes severe search time increase with the total number of combinations which, depend on \code{chain} and
substring lengths, and search window width. Slight performance improvement comes from variations in \code{rows} value
i.e. by balancing the number of combinations matrix \code{rows} with the total number of combinations matrices. The default
value is an acceptable starting point. Chains of several hundreds of characters, such as long protein chains, may take hours
on some machines. In such cases, \code{rows} value can be risen toward the order of 1e6. When \code{Y != NULL}, the partitioning
of chains runs asynchronously.

The \code{lower} and \code{upper} limits. Setting any or both of these arguments reduces the search time and only establishes
whether \emph{any} common substrings exist at all, without guaranteeing complete results. In case of very long chains, setting
these limits too wide apart may lead to memory allocation error. Checking the number of combinations before run is
recommended, see \link[RcppAlgos:combinatoricsCount]{RcppAlgos::comboCount}, \link[utils:combn]{utils::combn} etc.

NOTE: This utility uses background processing. Check "Security Considerations" in \strong{callr} package documentation.
}
\examples{

if (interactive()) {

# 1. A set of chains

 X = 'alpdxoipoyloiekladxoipoylyl'
 Y = 'kdxoipoylyydxoipoylopldxoipoylac'

# 1.1 Identify all 4-character common substrings
 system.time(a <- common(X, Y, 4))
 print(a)                                              # common substrings set

 # 1.2 Check the commonality of substrings in "a"
 chain = list(chain1 = 'alpdxoipoyloiekladxoipoylyl',
              chain2 = "kdxoipoylyydxoipoylopldxoipoylac")

 b = sapply(a, findLoc, chain, TRUE, TRUE, TRUE, simplify = FALSE)
 print(b)                                              # a named list


 any(lengths(b) == 0L)                                 # FALSE
 identical(length(a), length(b))                       # TRUE

# 2. Combinatorial Subset

 c = common(X, Y, from = 4, to = 9, lower = 5, upper = 100, outlist = TRUE)
 print(c)                                              # named list

# 3. Single Chain

 d = common(X, from = 3, to = 5)
 print(d)                                              # valid substrings set

}

}
\seealso{
\link[RcppAlgos]{comboIter}, \link[RcppAlgos]{comboGeneral}, \link[RcppAlgos]{comboCount}
}
\keyword{Proteomics}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/oneHotEncoder.R
\name{oneHot}
\alias{oneHot}
\title{One-hot Encoder And Decoder Of Variables}
\usage{
oneHot(x, type, omc = "dgCMatrix", verbose = TRUE)
}
\arguments{
\item{x}{a (named) vector or list for encoding. Missing data are removed. For decoding, a dense or sparse
matrix (preferably, the result of encoding) representing a single source data column}

\item{type}{symbol. Choices: \strong{encode} - one-hot encoding, \strong{decode} - revert to original}

\item{omc}{character length 1. \strong{O}utput \strong{m}atrix \strong{c}lass. Default, 'dgCMatrix', other
option, 'matrix'}

\item{verbose}{logical, default TRUE, display messages}
}
\value{
Encoding returns a matrix of \code{length(x)} rows and \code{length(unique(x))} columns or a warning.
Decoding returns a (named) vector or a warning. List vectors are returned unlisted. Integer(ish) vectors,
converted to integer, character vectors - to factor, double or logical vector types remain unchanged.
}
\description{
Encodes logical, categorical, integer and double type variables.
}
\details{
This utility one-hot encodes when \code{type = encode} and verifies the encoded result (or any
matrix of encodings obtained with \link{getEV} extractor) when \code{type = decode}. It detects illicit states.
}
\examples{

if (interactive()) {

# 1. Encode type "double"

x = runif(9)                           # numeric, length 9
names(x) = letters[1:9]                # named
typeof(x)
a = oneHot(x, encode)                  # a sparse matrix of "dgCMatrix" class
b = oneHot(a, decode)                  # a type "double" named numeric, length 9
isTRUE(all.equal(x, b))                # TRUE
typeof(b)
print(x); print(b)

# 2. Type "logical" with missing values

y = c(TRUE, TRUE, NA, FALSE, TRUE, NA) # logical, length 6 with missing values
typeof(y)
a = oneHot(y, encode, 'matrix')
print(a)                               # a dense matrix
b = oneHot(a, decode)                  # revert
all.equal(y, b)                        # missing values in y removed
typeof(b)
print(x); print(b)

# 3. iris data

data(iris)
a = lapply(iris, oneHot, encode)       # encode entire data
b = as.data.frame(
         lapply(a, oneHot, decode)     # revert
     )
identical(iris, b)                     # TRUE. Now, replace iris data with
                                       # mtcars data!

# 4. Illicit states in one-hot encoding

`3.41` = c(1,0,0,1,1,0,0,1)            # encoded type "double"
`0.12` = c(0,1,0,0,0,1,1,0)
 a = cbind(`3.41`, `0.12`)             # form encoded matrix
 print(a)                              # matrix resembling one-hot encoding
 x = oneHot(a, decode)                 # illicit state detected
 print(x)                              # list with 2 different data types

}

}
\keyword{conversion}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/akin-package.R
\docType{package}
\name{akin-package}
\alias{akin}
\alias{akin-package}
\title{Functional Utilities For Data Processing}
\description{
The intent of this package is utilitarian which explains the relatively large number of informative messages
displayed by most functions. Designed for large and very large data files, the package employs data.table,
list environments, sparse matrices, iterators, background and parallel processing in few places. Nevertheless,
Users could consider thread optimization or memoisation, as these techniques were outside the scope.

The package covers several areas of data processing: subset reading and writing of large data files, data tiling
(horizontal and vertical splitting) - suited for data conversion operations with local as well as global hold,
such as one-hot encoding - stratified, proportional, random or probabilistic data sampling, data normalization and
thresholding, substring location inside strings e.g. peptides inside protein chains, common substrings identification
and location-tabulation of amino acids, modifications or their associated monoisotopic masses inside modified peptides
covering various representations of protein mass spectrometry data with no pretense of exhaustiveness.

Comments and suggestions regarding tricky situations are provided however brief. Examples should be run individually
in R console. When possible, suppression of messages should be avoided.
}
\author{
\strong{Maintainer}: Dragos Bandur \email{dbandur@sympatico.ca}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tileD.R
\name{tileData}
\alias{tileData}
\title{Tile And Write Tiled Data To Disk}
\usage{
tileData(readpath, writepath = NULL, rows, splits, ...)
}
\arguments{
\item{readpath}{character length 1. Full path to the source file}

\item{writepath}{character length 1. Full path to the destination file}

\item{rows}{integer length 1. Number of rows in each subset. Internally, it determines the total
number of subsets before the vertical split}

\item{splits}{integer, length 1. Number of vertical data splits in each above subset. See \link{splitV}}

\item{...}{extra arguments to \link{splitH} e.g. \code{dropcols} for columns dropped from source data}
}
\value{
A \link[listenv]{listenv} of "data.table" class tiles. When \code{writepath} is given, it produces
a "rds" file containing data tiles.
}
\description{
Splits long and wide data files in lists of disjoint tiles for further processing.
}
\details{
Facilitates local operations on small size tiles by partitioning the data horizontally and vertically.
The list of tiles can be written to disk as "rds" file when a \code{writepath} destination is given. The written
data can then be read entirely or in subsets (see Example 2).

NOTE: This utility uses background processing. Check "Security Considerations" in \strong{callr} package documentation.
}
\examples{

if (interactive()) {

# Make a 'csv' file

data(iris)
tmpf = tempfile(fileext = '.csv')
write.table(iris, tmpf , sep = ',', row.names = FALSE, quote = FALSE)

# 1. Tile data

a = tileData(tmpf, rows = 10, splits = 3)     # 10x2 and 10x1 tiles
class(a)                                      # listenv, environment
str(a)                                        # nested list

tmpf1 = tempfile(fileext = '.rds')            # new location

# 2. Write tiled data

tileData(tmpf, tmpf1, rows = 10, splits = 3)
a = readRDS(tmpf1)[[1]]                  # partial read from new location
print(a)                                     # list component

unlink(tmpf)
unlink(tmpf1)
}


}
\seealso{
\link{splitH}, \link{splitV}, \link{tileHot}, \link[base]{readRDS}
}
\keyword{splitting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/akin-defunct.R
\name{akin-defunct}
\alias{akin-defunct}
\title{Defunct Functions In Package akin}
\description{
The following functions are defunct. Where available, alternative functions are suggested
at \code{help("akin-defunct")}.
}
\section{\code{common}}{
 this function is defunct. Use \link{fcommon} instead.
}

\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stratified_sample.R
\name{stratify}
\alias{stratify}
\title{Extract A Proportional Stratified Sample From A Data Set}
\usage{
stratify(
  X,
  target,
  stratum = NULL,
  size,
  thresh,
  seed = NULL,
  indx = TRUE,
  dis = NULL,
  args = list(),
  ext = FALSE,
  replace = FALSE,
  verbose = TRUE
)
}
\arguments{
\item{X}{any data array convertible to "data.table" class}

\item{target}{character length 1. The name of column considered to be the root stratum. For example, the name of the
'target' categorical column in a classification training set. This argument should always have a value}

\item{stratum}{character of length <= \code{ncol(data) - 1}. Default, NULL. Names of additional categorical data
columns which deepen the stratification}

\item{size}{integer length 1. Default, none. Value set by User. In this case, it is upper-bounded by the size of the
thinnest stratum having more than one row. Setting \code{size} value above this bound requires sampling with
replacement}

\item{thresh}{integer, length 1. Default, none. An automatic switch between sample size calculation formulae.
Can be set when \code{size} is missing from call. It can take as value any of the stratum thicknesses shown in the
output message

NOTE: it is recommended that \emph{both} \code{size} and \code{thresh} values are missing from call until
information on stratification becomes available after first run}

\item{seed}{integer length 1. Seed value for output reproducibility}

\item{indx}{logical. Default TRUE, returns the sample row index only. FALSE, returns the sampled data}

\item{dis}{symbol. Default NULL. One of the density or function \link[stats]{distributions} used for
generating probability vectors for probabilistic sampling}

\item{args}{list of arguments required by distributions as described in \link[stats:Distributions]{stats::distributions} documentation.
Default, none. \strong{NB} The list should \emph{never} include the first argument (\strong{x} or \strong{n})
required in documentation, as it is collected internally from each stratum

NOTE: Even if \code{seed} is set, the sample row index changes if either the distribution in \code{dis} or the
values in \code{args} is changed}

\item{ext}{logical, default FALSE. When TRUE, expands the output sampled data with the following extra columns:
\strong{row} - sample rows, \strong{strat} - stratum, \strong{n} - stratum total rows (i.e. thickness)
and \strong{size} - the sample size extracted from each stratum. Requires \code{indx = FALSE}}

\item{replace}{logical, default FALSE. When TRUE, sampling with replacement if \code{size} is present in call
and exceeds the thinnest stratum with more than one row}

\item{verbose}{logical, default TRUE, display messages}
}
\value{
A proportional or non-proportional stratified sample (depending on whether \code{size} is absent or present
in call), either as row index or as sampled data, compounded from random or probability samples taken from each
stratum. Informative messages are displayed. Existing data row names are preserved in the output case in which, the sampled
data output gains the column named "rn".
}
\description{
Obtains a proportional stratified sample from any data convertible to "data.table"
class containing categorical variables.
}
\details{
This utility is designed to find a true sample representation of the data under current stratification
by matching closely the proportionality of strata as long as argument \code{size} is missing from call.
Each distinct combination of \code{target} and \code{stratum} levels defines a stratum. For minimal
stratification, argument \code{target} must always have a value present in call. All one-row strata, when
formed, are simply appended to the compounded output.

\code{size}. As column in the extended output, it represents the size of the sample extracted from each
stratum, internally derived to be proportional to stratum thickness, unbounded by the thinnest stratum with more
than one row. Deep stratification along with high cardinality and imbalance may severely restrict the size of the
compounded output which is the sum of all stratum sizes plus the number of one-row strata. The sampling occurs at
stratum level except for one-row strata for which \code{size = 0} is interpreted as "no sampling".

As function argument, \code{size} is interpreted as the largest sample size without replacement that can be requested,
being bounded by the thinnest stratum with more than one row. The presence of \code{size} in call alters
the proportionality since each stratum - except one-row strata - contributes equally to the output size which is
the number of strata times the \code{size} value plus the number of one-row strata.

\code{thresh}. Automatic switch that modifies stratum sample size calculation method based on the extreme stratum
thickness values, stratification depth and total data rows. Internally, it searches for the formula that finds
at least one sample size accommodating the thinnest stratum with more than one row. Messages are displayed at runtime
although, in most cases the condition is satisfyed at first iteration. When \code{thresh >= nrow(data)}, each stratum
is sampled proportional with the ratio between thinnest and thickest strata, which may lead to a relatively small size
output. All other \code{thresh} values compromise slightly between output size and proportionality (see Example 3).
\subsection{Probabilistic Sampling}{

\code{dis}. The \code{prob} argument in \link[base:sample]{base::sample} cannot be used as required since the length of probability vector
varies with stratum thickness. Herein, stratum probability vectors are determined by the distribution specified in
argument \code{dis} which associates each stratum with a probability vector of thickness length. When \code{args} is
missing from call, \code{dis} uses the default argument values for respective distribution. An error is thrown when the
probability vector has insufficient number of non-zero values. See package \strong{stats}, "Distributions" documentation.

NOTE: The random variate generators i.e. the \strong{r*} version of \link[stats]{distributions}, generate vectors of absolute
\emph{random deviate} values which play the role of pseudo-probabilities conformant with the requirements listed in
\link[base:sample]{base::sample} documentation.
}
}
\examples{

if (interactive()) {

# 1. Row index for sampling

data(mtcars)
rowID = stratify(mtcars
               , target = 'cyl'
               , stratum = c('vs', 'am')
               , seed = 314)                                  # display information
print(rowID)                                                  # integer

# 2. Sampled data with extra-columns

smp = stratify(mtcars
            , 'cyl'
            , c('vs', 'am')
            , seed = 314
            , indx = FALSE
            , ext = TRUE)                                     # extra columns
print(smp)
identical(rowID, smp$row)                                     # TRUE

# 3. Impact of "thresh" value on output size

sl = list()
thresholds = c(2, 4, 12, 32)                                  # stratum thicknesses

for (t in seq(along=thresholds)) {
                  sl[[t]] = stratify(mtcars
                                  , 'cyl'
                                  , c('am', 'vs')
                                  , thresh = thresholds[t]
                                  , seed = 314
                                  , indx = FALSE, ext = TRUE)
                }
names(sl) = quote(thresholds)
print(sl)                                                     # stratified samples
                                                              # of various sizes

# 4. Probabilistic sampling

rowIDn = stratify(mtcars
             , 'cyl'
             , c('vs', 'am')
             , seed = 314
             , dis = pnorm                                    # Normal distribution
             , args = c(mean = 1, sd = 3))                    # no first argument!
rowIDb = stratify(mtcars
             , 'cyl'
             , c('vs', 'am')
             , seed = 314                                     # same seed
             , dis = pbeta                                    # Beta distribution
             , args = c(shape1 = 1, shape2 = 3))              # no first argument!

# Same seed but changing the distribution changes the sample row index
identical(rowIDn, rowIDb)                                     # FALSE

}

}
\seealso{
\link[base]{sample}, \link[stats]{distributions}
}
\keyword{sampling}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tileHot.R
\name{tileHot}
\alias{tileHot}
\title{One-hot Encoder Of Tiled Data}
\usage{
tileHot(readpath, rows, splits, omc = "dgCMatrix", ...)
}
\arguments{
\item{readpath}{character, length 1. Path to source data that is readable with \link[data.table:fread]{data.table::fread}}

\item{rows}{integer length 1. Number of rows in each data subset. Internally, it determines the total
number of subsets before the vertical split}

\item{splits}{integer, length 1. Number of vertical data splits in each subset, see \link{splitV}. Recommended for
very wide data frames. When \code{splits = 0}, no vertical splitting occurs}

\item{omc}{character length 1. \strong{O}utput \strong{m}atrix \strong{c}lass. Default, "dgCMatrix". Other option:
"matrix"}

\item{...}{reserved for \link{splitH} function arguments, such as \code{dropcols} or \code{orn = TRUE} which is needed for
single matrix output}
}
\value{
\itemize{
\item When \code{orn = FALSE}, an unnamed \link[listenv]{listenv} of sparse matrices. Recommended for very large
source data files. Before proceeding with list output, read NOTE 2 in \link{getEV} documentation. See Examples 1 and 2.
\item When \code{orn = TRUE}, a matrix.
}

NOTE 3: In this case, row and column binding operations were avoided to prevent situations described in NOTE 2, \link{getEV}
documentation. As result, the output matrix is gradually populated instead of being gradually expanded.

While \code{orn = TRUE} and \code{dropcols != NULL}:
\itemize{
\item When \code{omc = 'matrix'}, a \link[data.table]{data.table} containing encoded, as well as unencoded, dropped
columns placed in the leftmost positions.
\item When \code{omc = 'dgCMatrix'}, a two-component \link[listenv]{listenv}: a data table containing dropped,
unencoded columns and a sparse matrix containing the encoded columns. The row order in both components is identical.
See Examples below, and Example 2 in \link{getEV} documentation.
}

NOTE 4: In all above cases, specific encoded variables can be obtained with \link{getEV} extractor. When \code{orn = TRUE},
\link{oneHot} \emph{decoded} variables extracted from matrix outputs return named vectors having row numbers as names.
}
\description{
One-hot encodes tiled data.
}
\details{
This utility reads the data in disjoint subsets, tiles them and then one-hot encodes each tile. Encoded
tiles are returned as nested list of matrices, as a single matrix, as data frame or as a two-component data frame
and sparse matrix list, decided through combinations of \code{dropcols}, \code{omc} and \code{orn} values.

NOTE 1: traceability is assured by assembling the data as character names and values from columns marked for encoding.
As side effect, at run time the encoding is reported as being applied to "integer(ish)" values only with no loss in
accuracy. Empty source data columns gain the "NA" suffix and become single-column, single-valued matrices.

NOTE 2: this utility implements background, processing. Check "Security Considerations" in \strong{callr}
package documentation.
}
\examples{

if (interactive()) {

# 1. Shuffled data

tempf = tempfile(fileext = '.csv')
data(iris)
iris22 = iris[{ set.seed(327); sample.int(150) },]          # shuffled iris data
rownames(iris22) <- NULL                                    # remove shuffled row names
write.table(iris22, tempf, sep = ',', row.names = FALSE, quote = FALSE)

# 1.1 Output as List
# In most cases, list output requires shuffled data!

A = tileHot(readpath = tempf
          , rows = 14, splits = 3, print = FALSE)           # encoded data tiles
print(A)                                                    # a listenv
print(A[[1]])                                               # a snapshot

# 1.2 Retrieve iris22 data from encoded list output
X = sapply(names(iris22), \(n) getEV(A, n))                 # extract all encoded columns
Y = lapply(
         lapply(X, oneHot, decode)
                               , unname)                    # decoded columns are named vectors!
d = as.data.frame(Y)
identical(iris22, d)                                        # TRUE

unlink(tempf)

# 2. Unshuffled data

# Make unshuffled data 'csv' file
tempf = tempfile(fileext = '.csv')
write.table(iris, tempf, sep = ',', row.names = FALSE, quote = FALSE)

# 2.1 Output as list
# List output fails low cardinality variables on unshuffled data.

E = tileHot(readpath = tempf
              , rows = 14, splits = 3, print = FALSE)      # same as above

# 2.2 Retrieve iris data from encoded list output
V = sapply(names(iris), \(n) getEV(E, n))                  # warning
W = lapply(
         lapply(V, oneHot, decode)
                               , unname)                   # decoded columns are named vectors!
dd = as.data.frame(W)
identical(iris, dd)                                        # FALSE
all.equal(iris, dd)                                        # low cardinality "Species"

# 2.3 Output as matrix
# Matrix output handles low cardinality variables. No data shuffling required.

m = tileHot(readpath = tempf                               # low cardinality "Species"
          , rows = 14
          , splits = 3
          , orn = TRUE,                                    # needed for matrix output
          , print = FALSE)
print(m)                                                   # 150x126 sparse matrix

# 2.4 Retrieve iris data from encoded matrix output
P = sapply(names(iris), \(n) getEV(m, n))                  # extract encoded columns
Q = lapply(
           lapply(P, oneHot, decode)
                              , unname)                    # decoded columns are named vectors!
R = as.data.frame(Q)
identical(iris, R)                                         # TRUE

# 2.5 Output as "data.table" class
D = tileHot(readpath = tempf
          , rows = 14
          , splits = 3
          , omc = 'matrix'                                 # encoded dense matrix
          , dropcols = c('Petal.Width', 'Petal.Length')    # unencoded columns
          , orn = TRUE                                     # needed for matrix output
          , print = FALSE)
print(head(D, 10))                                         # a "data.table" class
dim(D)                                                     # 150x63

# 2.6 Output as a 2-component list
Dl = tileHot(readpath = tempf
          , rows = 14
          , splits = 3
          , omc = 'dgCMatrix'                              # the default class
          , dropcols = c('Petal.Width', 'Petal.Length')    # unencoded columns
          , orn = TRUE                                     # needed for matrix output
          , print = FALSE)
print(Dl)                                                  # 2-component listenv
print(Dl[[1]])                                             # unencoded columns
print(Dl[[2]])                                             # encoded sparse matrix

# iris data can be retrieved from the Dl list in similar fashion described above

unlink(tempf)

}

}
\seealso{
\link{splitH}, \link{splitV}, \link{oneHot}, \link[listenv]{listenv}, \link[Matrix]{Matrix}
}
\keyword{conversion}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modPep.R
\name{locateMod}
\alias{locateMod}
\title{Locate And Extract Modifications Or Monoisotopic Masses From A Modified Peptide}
\usage{
locateMod(string, wrap = "]", inbracket = ")", except = NULL, rmve = NULL)
}
\arguments{
\item{string}{character, length 1. Modified or unmodified peptide, or NULL}

\item{wrap}{character, length 1. The closing (right-hand) side of any of the bracket types ']', ')', '\}' that wrap
the modifications, such as in protein mass spectrometry data representation of modified peptides. Default, ']'}

\item{inbracket}{character, length 1. Same as above for brackets used inside modification wrappings. Default, ')'}

\item{except}{character, length >= 1. Default, NULL. Punctuation marks or characters that appear along modifications
and are needed to remain present in the output: '-', '+', ',', ';', ':', '=', '.', \verb{[:digit:]}, \verb{[:alpha:]}, '\\w+', ' '}

\item{rmve}{character, length 1. Default, NULL. Regular expression. Digits or extra characters that need to be removed from
the output (see Examples)}
}
\value{
A 'data.table' class data frame containing the unmodified peptide, the modified peptide, the modification site
(i.e. the amino acid code letter and location inside the peptide) and the associated modification(s). In case of monoisotopic
mass extraction, monoisotopic mass values populate column "Modification" as "character" types. Multiple modifications (identical
or not) found at the same site are listed as many times as they appear at that site. Unmodified, endogenous peptides are listed
with no other information. Empty strings are listed as such with a warning.
}
\description{
Finds and tabulates amino acid modification sites and extracts modifications or monoisotopic masses
from modified peptide data representation.
}
\details{
Although capable of handling most situations, it is recommended that the \code{wrap}ping bracket type
remains consistent throughout and the \code{inbracket} type \emph{be different} from \code{wrap}ping type.
No extra characters are removed from result when \code{except = rmve = NULL}.

This utility covers most data representation styles for modified peptide. However, clean data results are not guaranteed.
The template for letter casing accepted for modified peptide and for modifications should match those presented in Examples:
upper case for peptide and mixed case for modifications.
}
\examples{

if (interactive()) {

# Completely made-up modified peptides:

# 1. Modifications

# 1.1 Default brackets
string = 'K[Prop_A][Met][Prop (C)]PSSABCELR[Prop][Prop][Prop]FQC[Carba (C)]GQQ[Met +44]TARP'

a = locateMod(string)
print(a)                                                              # with extra-characters
b = locateMod(string, except = '\\\\w+', rmve = '(\\\\(.*\\\\)|_[A-Z]|[0-9])')
print(b)                                                              # without extra-characters

# In this example argument "rmve" contains the default in-brackets

# 1.2 Alternative bracketing

string = 'K{Prop_A}{Met}{Prop [A]}PSSABCELR{Prop +15}{Prop}{Prop}FQC{Carba [C]}GQQ{Met +44}TARP'

c = locateMod(string, '}', ']')
print(c)
d = locateMod(string, '}', ']', except = '\\\\w+', rmve = '(\\\\[.*\\\\]|_[A-Z]|[0-9])')
print(d)

# In this example argument "rmve" contains the alternative in-brackets

# 2. Empty string

empty = locateMod(""); print(empty)

# 3. Monoisotopic masses

string = 'TAAC[+57.021464]PPC[+57.021464]PAPPAPS[+162.052824]VFLTLMISR'
e = locateMod(string)
print(e)                                                             # with extra-characters
f = locateMod(string, rmve = '[[:punct:]]')$Modification
print(f)                                                             # incorrect values
g = locateMod(string, rmve = '\\\\+')$Modification
print(g)                                                             # correct!
class(g)                                                             # character
}

}
\seealso{
\link[base]{regex}
}
\keyword{Proteomics}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/findLoc.R
\name{findLoc}
\alias{findLoc}
\title{Find Substring Locations Inside A String}
\usage{
findLoc(
  subchain,
  chain,
  outlist = FALSE,
  named = FALSE,
  all. = FALSE,
  which = min,
  ignore.case = TRUE,
  perl = FALSE,
  fixed = FALSE,
  useBytes = FALSE
)
}
\arguments{
\item{subchain}{character, length 1, e.g. a peptide sequence}

\item{chain}{(named) character, length 1 or a (named) list of such characters such as a
list of protein chains obtained from a \emph{fasta} file}

\item{outlist}{logical. Default, FALSE, the output is a (named) integer vector of locations. Otherwise,
it is a (named) list of location vectors, each corresponding to a chain in a list of chains}

\item{named}{logical. Default, FALSE. Output is not named. Otherwise, the output is named}

\item{all.}{logical, default FALSE, returns the leftmost or the rightmost location inside the chain.
When TRUE, returns all locations inside the chain for each chain in a list}

\item{which}{symbol. Location to report. Default, \strong{min}. Requires \code{all. = FALSE}. \code{which = min}
returns the leftmost location inside the chain, e.g. closest to the N-terminus of a protein chain. \code{which = max}
returns the rightmost location, e.g. closest to the C-terminus of a protein chain. Overwritten when \code{all. = TRUE}}

\item{ignore.case, perl, fixed, useBytes}{arguments to \link[base:grep]{base::gregexpr}}
}
\value{
A (named) integer or a (named) list of integer vectors of \code{subchain} locations inside the \code{chain}.
}
\description{
Finds all locations of a known character substring inside a character string.
}
\details{
Wrapper to \link[base:grep]{base::gregexpr}, the function scans all chains in a list of chains to find subchain locations.
The location is defined as the position inside the chain relative to the left end of the chain of the first subchain character.
}
\examples{

if (interactive()) {

# 1. List of chains

chain = list(chain1 = 'alpdxoipoyloiekladxoipoylyl',
             chain2  = 'kdxoipoylyydxoipoylopldxoipoylac')
subchain = 'DXOIPOYL'                                                  # ignoring the case

findLoc(subchain, chain, outlist = TRUE, named = TRUE, all. = TRUE)    # named list
findLoc(subchain, chain, named = TRUE, all. = TRUE)                    # named integer
findLoc(subchain, chain, outlist = TRUE, named = TRUE)                 # the leftmost positions
findLoc(subchain, chain, which = max, named = TRUE)                    # the rightmost positions

# 2. Single chain

chain = chain[[1]]

findLoc(subchain, chain, all. = TRUE)
findLoc(subchain, chain, which = max)
findLoc(subchain, chain)                                               # default location
findLoc(subchain, chain, which = max, ignore.case = FALSE)             # not ignoring the case

}

}
\seealso{
\link[base]{gregexpr}
}
\keyword{Proteomics}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/seqRead.R
\name{splitH}
\alias{splitH}
\title{Read Or Write Subsets Of Data Files From Or To Disk}
\usage{
splitH(readpath, writepath = NULL)
}
\arguments{
\item{readpath}{character length 1. Full path to the source file}

\item{writepath}{character length 1. Full path to the destination file}
}
\value{
At stage 1, displayed information and a function (a closure). At stage 2, a "data.table" class subset
of data or a printout of said subset when written on disk.
}
\description{
Reads or writes data files from/to disk in disjoint subsets. This is a two-stage function (see Examples).
}
\details{
Above arguments apply to Stage 1 only. The arguments for Stage 2 function, which is the output of Stage 1,
are the following:

\code{rows} integer, length 1. Number of rows per subset. When rows = Inf, the data can be either
copied \emph{as is} or moved to a new location

\code{seq} logical, default TRUE: read discrete subsets. Otherwise, progressively appended subsets
from first to current

\code{dropcols} character of \code{length < ncol(data)}. Columns to drop. Works only when \code{rows} is
finite. Replaces argument \code{select} from \link[data.table:fread]{data.table::fread}

\code{how} symbol. Works only when \code{rows = Inf} and \code{writepath} location is given.
Options: \code{how = scp}, data file is copied \emph{as is} to \code{writepath} location;
\code{how = mv}, data file is moved to \code{writepath} location

\code{print} logical, default TRUE, each subset written to disk is shown in console. Setting print to
FALSE could increase writing speed

\code{orn} logical, default FALSE. When TRUE, the \strong{o}riginal data \strong{r}ow \strong{n}umbers
are shown in each subset

The main purpose of this utility is to bring manageable subsets from very large data into the working
environment for further processing when \code{writepath = NULL}. When \code{orn = TRUE}, each subset
receives a new column named "srn" showing \strong{s}ource data \strong{r}ow \strong{n}umbers. This column
is absent from subsets written to disk regardless of \code{orn} value. The source data file can be any
type of file readable by \link[data.table:fread]{data.table::fread}.

At the first stage:
\itemize{
\item the utility retrieves information about source data without loading them into memory and also provides
the new function which, in the second stage:
\item reads source data in successive disjoint subsets (\code{rows < Inf}) and brings them into the work
environment (writepath = NULL), \strong{or}
\item writes subsets to \code{writepath} location appending them automatically to the destination file.
During writing, if (\code{print = TRUE}) the displayed subsets are just printouts (class "NULL"). When
\code{writepath = NULL}, displayed subsets are objects.
}

There is a functional difference between \code{rows = Inf} and \code{rows = nrow(data)}:
\itemize{
\item when \code{rows = Inf}, the size of source data is irrelevant. They can be either copied (\code{how = scp})
or moved (\code{how = mv}) to \code{writepath} destination without being loaded into memory.
\item when \code{rows} has finite value, the size of source data is relevant and data columns can be dropped.
}
}
\examples{

if (interactive()) {

# Make a 'csv' file

data(mtcars)
tmpf = tempfile(fileext = '.csv')
write.table(mtcars,tmpf , sep = ',', row.names = FALSE, quote = FALSE)

# 1. Read data file step by step

# 1.1 Get information on data
r = splitH(readpath = tmpf)                             # stage 1
class(r)                                                # function

# 1.2 Read data iteratively                             # stage 2
a = r(rows = 11, dropcols = c('am', 'vs'))              # iter1 no original row numbers
b = r(rows = 11, dropcols = c('am', 'vs'), orn = TRUE)  # iter2 w. original row numbers
c = r(rows = 11, dropcols = c('am', 'vs'))              # iter3 the last subset
d = r(rows = 11, dropcols = c('am', 'vs'))              # iter4 stop! Return to stage 1
print(list(a, b, c))

# 2. Read data file completely

r = splitH(readpath = tmpf)                             # stage 1
n = ceiling(32/13)                                      # read 13 rows at a time
a = replicate(n, r(rows = 13), simplify = FALSE)        # read file
class(a)                                                # list
print(a)                                                # a list of tables

tmpf1 = tempfile(fileext = '.csv')                      # new location

# 3. Iteratively write to new location

r = splitH(readpath = tmpf, writepath = tmpf1)          # stage 1
n = ceiling(32/11)                                      # 11 rows each time
invisible(
 replicate(n, r(rows = 11) , simplify = FALSE)          # write to new location
 )
a = data.table::fread(tmpf1)                            # check result
dim(a)
print(head(a))

unlink(tmpf1)

tmpf2 = tempfile(fileext = '.csv')                      # new location

# 4. Move file from tmpf to another location

r = splitH(readpath = tmpf, writepath = tmpf2)          # stage 1
r(rows = Inf, how = mv, print = FALSE)                  # move to new location
a = data.table::fread(tmpf2)                            # check result
print(head(a))

unlink(tmpf)
unlink(tmpf2)

}

}
\references{
Part of internal code for Stage 1 was inspired by
\href{https://github.com/Rdatatable/data.table/issues/7169}{data.table Issue# 7169}
}
\seealso{
Linux commands \strong{scp} and \strong{mv}
}
\keyword{splitting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cover.R
\name{cover}
\alias{cover}
\title{Calculate The Number Of Combinations Of String Covering}
\usage{
cover(x, valid. = FALSE, wplot = FALSE)
}
\arguments{
\item{x}{integer, length = 1 or character vector, length > 1, or character string}

\item{valid.}{logical, default, FALSE. When TRUE, valid substrings are returned. Requires character or string}

\item{wplot}{logical, FALSE. When TRUE, displays a plot of combinations versus number of characters in string}
}
\value{
When \code{valid. = FALSE} a prettified named vector showing the total, minimum and maximum number of
combinations for respective string covering. When \code{valid. = TRUE}, a character vector of valid substrings
longer than 2 characters each. In both cases, a plot is returned when \code{wplot = TRUE}.
}
\description{
Calculates the cardinality of substring family (\emph{the covering}) and returns valid elements.
}
\details{
This function finds the total, the minimum and the maximum number of combinations of individual string
covering or returns a subset of valid substrings (i.e. elements of the string) by retaining even transpositions
inside the combinations matrix.

NOTE 1: Herein, the \emph{covering} represents the substring family that preserves the order of characters inside
the string. Hence, the use of combinations instead of partial permutations, \code{n!/(n-r)!}, as the order
inside each selection group \code{r} is irrelevant.

NOTE 2: Very long strings may result in memory allocation error when \code{valid. = TRUE}. Use \link{fcommon} instead.

NOTE 3: When \code{x} is an integer associated with an existing string, \code{x} must equal either \code{nchar(string)}
or \code{length(character)}.
}
\examples{

if (interactive()) {

# 1. x is a character

# 1.1 A string
x = 'tyrvvsvltvlhqdwlngkeykck'

a = cover(x)
print(a)

# 1.2 A vector
x = strsplit(x, '')[[1]]

b = cover(x)
print(b)

# 2. An integer

n = length(x)
c = cover(n)
print(c)

# 3. Valid substrings with plot

d = cover(x, valid. = TRUE, wplot = TRUE)
print(head(d, 30))

}


}
\seealso{
\link{fcommon}
}
\keyword{Proteomics}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/splitLONG.R
\name{splitV}
\alias{splitV}
\title{Split Data Vertically Into Subsets}
\usage{
splitV(data, splits)
}
\arguments{
\item{data}{a "data.table" class data frame or convertible to "data.table" class}

\item{splits}{integer, length 1, of value <= \code{ncol(data)}. The number of disjoint
selections (i.e. vertical splits). When \code{splits = 0}, there are no splits}
}
\value{
An exhaustive \link[listenv]{listenv} of disjoint column groups from original data
}
\description{
Splits data into unequal and disjoint groups of columns (i.e. vertical splits)
}
\details{
The smaller the \code{splits} value, the wider the column groups. Column order from
source data is not preserved
}
\examples{

if (interactive()) {

# 1. Split iris data vertically

data(iris)
a = splitV(iris, splits = 3)     # split data in 3 column groups
class(a)                         # listenv, environment
print(as.list(a))                # list

}
}
\keyword{splitting}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fcommon.R
\name{fcommon}
\alias{fcommon}
\title{Fast Identify Common Substrings In A Pair Of Strings}
\usage{
fcommon(x, y)
}
\arguments{
\item{x, y}{character, length 1 each: a string, such as a protein chain. \code{y} can be missing}
}
\value{
A sorted character vector of common substrings of length >= 2 characters each. When \code{y} is missing from call, a sorted
character vector of valid substrings in \code{x} of length >= 2 characters each.
}
\description{
Checks and identifies substrings that are common to a pair of strings.
}
\details{
This utility identifies common substrings in the \code{x}, \code{y} pair of strings by isolating \emph{sequences} of
identical characters in both strings which then, are packed into substrings and validated. All one-character substrings are removed.
When \code{y} is missing, \code{x} is cleaved at each letter producing all substrings longer than 2 characters.
}
\examples{

if (interactive()) {

 # 1. Check for common substrings in the pair below

 x = 'dvvmtqsplslpvtpgepasiscrssqslaktyrvvsvltvlhqdwlngkeykckvv'
 y = 'mtqspltyrvvsvltvlhqdwlngkeykcksnkalpapiektisk'

# 1.1 Common substrings
 system.time(a <- fcommon(x, y))
 print(head(a, 30))

# 1.2 Cleaving (slow on very long strings!)
 system.time(aa <- fcommon(x))
 system.time(bb <- fcommon(y))

# 1.3 Identical results
 A = sort(intersect(aa, bb))                                # common substrings
 identical(a, A)                                            # TRUE

# 2. Different methods for valid substrings

x = 'tyrvvsvltvlhqdwlngkeykck'

# 2.1. Combinations matrix (slower!)
system.time(am <- cover(x, valid. = TRUE))                  # valid substrings

# 2.2 String cleaving
system.time(ac <- fcommon(x))                               # valid substrings

identical(am, ac)                                           # TRUE

}

}
\seealso{
\link{cover}
}
\keyword{Proteomics}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/normalizer.R
\name{score}
\alias{score}
\title{Scaling And Thresholding Of Numeric Variables}
\usage{
score(x, how, filter = NULL, ...)
}
\arguments{
\item{x}{numeric vector, length > 1. Variable to be scaled or filtered}

\item{how}{symbol. Choices are \strong{range}, \strong{stdev} or \strong{relu}}

\item{filter}{character, length 1. Default NULL. Choices "positive", "negative". Requires \code{how = relu}}

\item{...}{list reserved for User input of paired values, statistic or otherwise. The list uses individual
ellipsis arguments therefore, order of values needs be respected at all times e.g. when \code{how = range}  min value
\emph{first}, max value \emph{second}. When \code{how = stdev}, mean value \emph{first}, standard deviation \emph{second}.}
}
\value{
Numeric. When missing \code{...} and \code{how != relu}, scaled values using \code{x} own sample statistics.
Otherwise, scaling is based on values inputted by User. When \code{how = relu}, \code{x >= 0} or \code{x <= 0}, depending
on \code{filter} setting.
}
\description{
Implements classical methods for data scaling: range, z-score normalization, location and location-scale
normalization, as well as data thesholding through the simplest form of ReLU rectifier. Missing values are removed in
all cases.
}
\details{
Normalization (scaling) can be applied locally on subsets of \code{x} when User inputs the values in the \code{...} list.
Otherwise, the scaling is global i.e. it is applied to \code{x} as a whole. No assumptions regarding the underlying distribution
of \code{x} are made.

\code{how != relu}. When \code{...} is empty, the function uses the sample statistics of \code{x} e.g. the mean, range
or standard deviation. Otherwise, it uses values inputted by User case in which, \emph{location-scale} normalization
requires \code{how = stdev} and the \code{...} list filled as follows: the \code{min(x)} \emph{or} \code{max(x)}, or any other
value first and \code{sd(x) > 0} or any other positive value second. In particular, \emph{location} normalization works similarly
but with the second value = 1. Other location types, e.g. x/max(x), are obtainable.

NOTE: when \code{...} is populated with custom values, all other arguments should be present and named (see Examples).

\code{how = relu}. This option acts as numeric thresholding locally as well as globally. It stands for \strong{re}ctified
\strong{l}inear \strong{u}nit and involves no statistics. It applies to numeric types that have ordering property (double,
integer). On return, all \code{x} attributes are dropped. When \code{filter = 'positive'}, all negative values are set to zero
while positive values remain unchanged. Alternatively, when \code{filter = 'negative'}, all negative values remain unchanged
while all other values are set to zero. The "negative" option was added for symmetry.
}
\examples{

if (interactive()) {

# 1. ReLU thresholding

x = { set.seed(223); sort(runif(10, -3, 3)) }
y = score(x, relu, 'positive'); y
z = score(x, relu, 'negative'); z

# 1.1 ReLU Plot
olp = par(no.readonly = TRUE)
par(list(mar = c(1,1,1,1), mgp = c(0,0,0), tcl = -0.01, pty = 's'))
plot(x, y, type = 'l', col = 'steelblue', lwd = 2 ,
       xlim = c(min(x), max(x)), ylim = c(min(x), max(x))
     , ylab = expression(ReLU(x)), xaxs = 'i', yaxs = 'i', axes = FALSE, cex.lab = 0.7)
axis(1, pos = 0, cex.axis = 0.6) ; axis(2, pos = 0, cex.axis = 0.6)
points(x, z, type = 'l', col = 'orangered', lwd = 2)
legend('topleft', legend = c('positive', 'negative'),
       col = c('steelblue', 'orangered'), pch = 'l', lwd = 2, cex = 0.6, bty = 'n')
par(olp)

# 2. Location and location-scale

# 2.1 Location (e.g. "x - max(x)")
x = 1:10
M = max(x)
std = 1
a = score(x, stdev, NULL, M, std); a

# 2.2 Location (e.g. "x/max(x)")
m = 0                                                            # the mean
M = max(x)                                                       # or any value
b = score(x, range, NULL, m, M); b

# 2.3 Location-scale (e.g. "(x - max(x))/sd(x)")
M = max(x)                                                       # or any value
std = sd(x)                                                      # or any value > 0
c = score(x, stdev, NULL, M, std); c

# m, M and std above can be replaced with any values decided by User

# 3. Classical normalization

# 3.1 Range
d = score(x, range); d

# 3.2 z-score
e = score(x, stdev); e

# 4. Local vs. global z-score normalization

data(mtcars)
  x = mtcars$wt
  m = mean(x)
std = sd(x)

ll = split(x, f = as.factor(mtcars$cyl))        # partitioned x

# 4.1 Local scaling
aa = lapply(ll, score, stdev, NULL, m, std)     # filled ... list
na = unlist(aa, FALSE, FALSE)

# 4.2 Global scaling
nb = score(x, stdev)

# 4.3 Local as well as global hold
identical(sort(na), sort(nb))                   # TRUE

}

}
\references{
\href{https://en.wikipedia.org/wiki/Ancillary_statistic}{Ancillary
Statistic} for location and location-scale distributions
}
\keyword{thresholding}
